"""
Script for statistically evaluating various aspects of tsinfer performance.
"""
import argparse
import sys
import collections
import random
import concurrent.futures
import time
import warnings
import os.path
import colorsys
import json

import numpy as np
import pandas as pd
import matplotlib as mp
# Force matplotlib to not use any Xwindows backend.
mp.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import collections as mc
import seaborn as sns
import tqdm
import scipy.stats
import colorutils


import tsinfer
import tsinfer.cli as cli
import msprime

def make_errors(v, p):
    """
    For each sample an error occurs with probability p. Errors are generated by
    sampling values from the stationary distribution, that is, if we have an
    allele frequency of f, a 1 is emitted with probability f and a
    0 with probability 1 - f. Thus, there is a possibility that an 'error'
    will in fact result in the same value.
    """
    w = np.copy(v)
    if p > 0:
        m = v.shape[0]
        frequency = np.sum(v) / m
        # Randomly choose samples with probability p
        samples = np.where(np.random.random(m) < p)[0]
        # Generate observations from the stationary distribution.
        errors = (np.random.random(samples.shape[0]) < frequency).astype(int)
        w[samples] = errors
    return w


def generate_samples(ts, error_p):
    """
    Returns samples with a bits flipped with a specified probability.

    Rejects any variants that result in a fixed column.
    """
    G = np.zeros((ts.num_sites, ts.num_samples), dtype=np.int8)
    for variant in ts.variants():
        done = False
        # Reject any columns that have no 1s or no zeros
        while not done:
            G[variant.index] = make_errors(variant.genotypes, error_p)
            s = np.sum(G[variant.index])
            done = 0 < s < ts.sample_size
    return G


def infer_from_simulation(
        ts, recombination_rate=1, input_error=0, sample_error=0, engine="C",
        path_compression=True):
    if input_error == 0:
        genotypes = ts.genotype_matrix()
    else:
        genotypes = generate_samples(ts, input_error)
    positions = [mut.position for mut in ts.mutations()]
    return tsinfer.infer(
        genotypes, positions=positions, sequence_length=ts.sequence_length,
        recombination_rate=recombination_rate, sample_error=sample_error,
        engine=engine, path_compression=path_compression)


def get_mean_rf_distance(ts1, ts2):
    """
    Returns the mean distance between the trees in the specified tree sequences.
    """
    assert ts1.sample_size == ts2.sample_size
    assert ts1.sequence_length == ts2.sequence_length
    trees1 = []
    intervals1 = []
    trees2 = []
    intervals2 = []
    tns = dendropy.TaxonNamespace()
    for t in ts1.trees():
        dt = dendropy.Tree.get(data=t.newick(), schema="newick", taxon_namespace=tns)
        trees1.append(dt)
        intervals1.append(t.interval)
    assert len(trees1) == ts1.num_trees
    for t in ts2.trees():
        dt = dendropy.Tree.get(data=t.newick(), schema="newick", taxon_namespace=tns)
        trees2.append(dt)
        intervals2.append(t.interval)
    assert len(trees2) == ts2.num_trees
    j1 = 0
    j2 = 0
    total_distance = 0
    total_metric = 0
    # I haven't tested this algorithm thoroughly, so there might be corner cases
    # not handled correctly. However, the total_distance assert below should
    # catch the problem if it occurs.
    while j1 < len(trees1) and j2 < len(trees2):
        # Each iteration of this loop considers one overlapping interval and
        # increments the counters.
        l1, r1 = intervals1[j1]
        l2, r2 = intervals2[j2]
        l = max(l1, l2)
        r = min(r1, r2)
        rf_distance = dendropy.calculate.treecompare.symmetric_difference(
                trees1[j1], trees2[j2])
        total_metric += rf_distance * (r - l)
        total_distance += r - l
        if r1 <= r2:
            j1 += 1
        if r1 >= r2:
            j2 += 1
    # assert total_distance, ts1.sequence_length)
    return total_metric / total_distance


def check_basic_performance():
    # Basic check to ensure that we get the same results regardless of the recombination
    # rate

    num_samples = 10
    MB = 10**6
    sites = []
    trees = []
    edges = []
    nodes = []
    rf_distance = []
    recombination_rate = []
    for seed in range(1, 10):
        ts_source = msprime.simulate(
            num_samples, length=1*MB, Ne=10**4, recombination_rate=1e-8, mutation_rate=1e-8,
            random_seed=seed)
        print("sim: n = ",
                ts_source.num_samples, ", m =", ts_source.num_sites,
                "num_trees = ", ts_source.num_trees)
        for exponent in range(0, 6):
            infer_recomb_rate = 10**(-exponent)
            ts_inferred = infer_from_simulation(
                ts_source, recombination_rate=infer_recomb_rate, sample_error=0)
            recombination_rate.append(infer_recomb_rate)
            rf = get_mean_rf_distance(ts_source, ts_inferred)
            rf_distance.append(get_mean_rf_distance(ts_source, ts_inferred))
            sites.append(ts_source.num_sites)
            trees.append(ts_inferred.num_trees / ts_source.num_trees)
            nodes.append(ts_inferred.num_nodes / ts_source.num_nodes)
            edges.append(ts_inferred.num_edges / ts_source.num_edges)

            # print("recombination_rate = ", recombination_rate)
            # print("mean rf = ", rf)
            # print("num_trees = ", ts_inferred.num_trees / ts_source.num_trees)
            # print("num_edges= ", ts_inferred.num_edges/ ts_source.num_edges)
    df = pd.DataFrame(data={
        "num_samples": num_samples,
        "recombination_rate": recombination_rate,
        "sites": sites,
        "trees": trees,
        "edges": edges,
        "nodes": nodes,
        "rf_distance": rf_distance})
    for y_value in ["trees", "nodes", "edges", "rf_distance"]:
        plt.figure()
        plot = sns.boxplot(x=df["recombination_rate"], y=df[y_value])
        plt.savefig("tmp__NOBACKUP__/{}.png".format(y_value))


def check_effect_error_param(input_error=0.0):
    num_samples = 10
    MB = 10**6
    sites = []
    trees = []
    edges = []
    nodes = []
    rf_distance = []
    error = []
    for seed in range(1, 10):
        ts_source = msprime.simulate(
            num_samples, length=1*MB, Ne=10**4, recombination_rate=1e-8, mutation_rate=1e-8,
            random_seed=seed)
        print("sim: n = ",
                ts_source.num_samples, ", m =", ts_source.num_sites,
                "num_trees = ", ts_source.num_trees)
        for err in [0] + list(10.0**-np.arange(1, 5)):
            ts_inferred = infer_from_simulation(
                ts_source, recombination_rate=1, input_error=input_error, sample_error=err)
            error.append(err)
            rf = get_mean_rf_distance(ts_source, ts_inferred)
            rf_distance.append(get_mean_rf_distance(ts_source, ts_inferred))
            sites.append(ts_source.num_sites)
            trees.append(ts_inferred.num_trees / ts_source.num_trees)
            nodes.append(ts_inferred.num_nodes / ts_source.num_nodes)
            edges.append(ts_inferred.num_edges / ts_source.num_edges)
        df = pd.DataFrame(data={
            "num_samples": num_samples,
            "error": error,
            "sites": sites,
            "trees": trees,
            "edges": edges,
            "nodes": nodes,
            "rf_distance": rf_distance})
        df.to_csv("tmp__NOBACKUP__/error_{}.csv".format(input_error))
        # df = pd.read_csv("tmp__NOBACKUP__/error_{}.csv".format(input_error))
        plt.figure()
        fig, axs = plt.subplots(nrows=2, ncols=2)
        for j, y_value in enumerate(["trees", "nodes", "edges", "rf_distance"]):
            sns.boxplot(x=df["error"], y=df[y_value], ax=axs[j // 2, j % 2])
        plt.suptitle("input_error = {}".format(input_error))
        plt.savefig("tmp__NOBACKUP__/error_{}.png".format(input_error))
        plt.clf()


def check_variable_recomb():
    rate = 1e-10
    Mb = 10**6
    num_samples = 20
    seed = 10
    recomb_map = msprime.RecombinationMap(
        positions=[0, 1*Mb, 1.1 * Mb, 2 * Mb],
        rates = [rate, 100 * rate, rate, 0], num_loci=10000)
    sites = []
    trees = []
    edges = []
    nodes = []
    rf_distance = []
    flat_rate = []
    for seed in range(1, 100):
        ts_source = msprime.simulate(
            num_samples, recombination_map=recomb_map, Ne=10**4, mutation_rate=1e-8,
            random_seed=seed,)
        print(
            "sim: n = ", ts_source.num_samples, ", m =", ts_source.num_sites,
            "num_trees = ", ts_source.num_trees)
        samples = np.zeros((ts_source.num_samples, ts_source.num_sites), dtype=np.int8)
        for variant in ts_source.variants():
            samples[:, variant.index] = variant.genotypes
        positions = np.array([site.position for site in ts_source.sites()])
        genetic_positions = np.array([
            recomb_map.physical_to_genetic(x) for x in positions])
        recombination_rate = np.zeros(ts_source.num_sites)
        recombination_rate[1:] = genetic_positions[1:] - genetic_positions[:-1]
        recombination_rate /= recomb_map.get_num_loci()

        for recomb_rate in [1, recombination_rate]:
            ts_inferred = tsinfer.infer(
                samples=samples, positions=positions,
                sequence_length=ts_source.sequence_length,
                recombination_rate=recomb_rate, sample_error=0)
            rf = get_mean_rf_distance(ts_source, ts_inferred)
            rf_distance.append(get_mean_rf_distance(ts_source, ts_inferred))
            flat_rate.append(recomb_rate is 1)
            sites.append(ts_source.num_sites)
            trees.append(ts_inferred.num_trees / ts_source.num_trees)
            nodes.append(ts_inferred.num_nodes / ts_source.num_nodes)
            edges.append(ts_inferred.num_edges / ts_source.num_edges)
        df = pd.DataFrame(data={
            "num_samples": num_samples,
            "sites": sites,
            "trees": trees,
            "edges": edges,
            "nodes": nodes,
            "rf_distance": rf_distance,
            "flat_rate": flat_rate})
            # df.to_csv("tmp__NOBACKUP__/error_{}.csv".format(input_error))
    plt.figure()
    fig, axs = plt.subplots(nrows=2, ncols=2)
    for j, y_value in enumerate(["trees", "nodes", "edges", "rf_distance"]):
        sns.boxplot(x=df["flat_rate"], y=df[y_value], ax=axs[j // 2, j % 2])
    plt.suptitle("variable recomb")
    plt.savefig("tmp__NOBACKUP__/variable_recomb.png")
    plt.clf()

    # print(ts_source.num_sites)
    # print(ts_source.num_trees)
    # for t in ts_source.trees():
    #     print(t.interval[1] - t.interval[0])


# TODO these are really unit tests. Move them into the tests directory.
def check_single_tree_one_mutation_per_branch():
    for num_samples in [4]:
        num_samples = 10
        for seed in range(1, 10):
            ts_source = msprime.simulate(num_samples, length=2 * num_samples, random_seed=seed)
            # Put a mutation on every branch.
            t = ts_source.dump_tables()
            for u in range(ts_source.num_nodes - 1):
                t.sites.add_row(position=u, ancestral_state="0")
                t.mutations.add_row(site=u, node=u, derived_state="1")
            ts_source = msprime.load_tables(**t.asdict())
            ts_inferred = infer_from_simulation(ts_source)
            assert ts_inferred.num_trees == 1


def verify_trees_equal(ts_source, ts_inferred):
    """
    Verifies that the inferred trees are topologically equal to the specified
    inferred trees.
    """
    for t in ts_source.trees():
        print(t.interval)
        print(t.draw(format="unicode"))
    print("++")
    for t in ts_inferred.trees():
        print(t.interval)
        print(t.draw(format="unicode"))

    print("==")


# TODO these are really unit tests. Move them into the tests directory.
def check_many_trees_one_mutation_per_branch():
    # for num_samples in [10, 100, 1000, 10000]:

    import daiquiri
    daiquiri.setup(level="DEBUG", outputs=(daiquiri.output.Stream(sys.stdout),))
    num_samples = 4
    print("num_samples = ", num_samples)
    # for seed in range(1, 1000):
    for seed in [19]:
        print(seed)
        recombination_map = msprime.RecombinationMap.uniform_map(
                100, 0.002, num_loci=100)
        ts_source = msprime.simulate(
                num_samples, recombination_map=recombination_map,
                random_seed=seed, model="smc_prime")
        # Put a mutation on every branch.
        tables = ts_source.dump_tables()
        j = 0
        for tree in ts_source.trees():
            left, right = tree.interval
            n = len(list(tree.nodes()))
            delta = (right - left) / n
            x = left
            for u in tree.nodes():
                if u != tree.root:
                    tables.sites.add_row(position=x, ancestral_state="0")
                    tables.mutations.add_row(site=j, node=u, derived_state="1")
                    j += 1
                    x += delta
        print(tables)
        ts_source = msprime.load_tables(**tables.asdict())
        print("samples")
        print(ts_source.genotype_matrix().T)
        ts_inferred = infer_from_simulation(ts_source,
                path_compression=False,
                # engine="Py-matrix")
                engine="P")
        print("num_trees", ts_source.num_trees, ts_inferred.num_trees)
        print("num_edges", ts_source.num_edges, ts_inferred.num_edges)
        verify_trees_equal(ts_source, ts_inferred)
        print(ts_inferred.tables)
        assert ts_source.num_trees >= ts_inferred.num_trees
        assert ts_source.num_edges >= ts_inferred.num_edges


# TODO these are really unit tests. Move them into the tests directory.
def check_single_tree_one_mutation_per_branch():
    for num_samples in [10, 100]:
        print("num_samples = ", num_samples)
        for seed in range(1, 10):
            ts_source = msprime.simulate(num_samples, length=2 * num_samples, random_seed=seed)
            # Put a mutation on every branch.
            t = ts_source.dump_tables()
            for u in range(ts_source.num_nodes - 1):
                t.sites.add_row(position=u, ancestral_state="0")
                t.mutations.add_row(site=u, node=u, derived_state="1")
            ts_source = msprime.load_tables(**t.asdict())
            ts_inferred = infer_from_simulation(ts_source)
            assert ts_inferred.num_trees == 1

def check_single_tree_two_mutations_per_branch():
    for num_samples in [10, 100]:
        print("num_samples = ", num_samples)
        num_mutations = 2 * num_samples * 2
        d = 1 / num_mutations
        for seed in range(1, 10):
            ts_source = msprime.simulate(num_samples, length=num_mutations, random_seed=seed)
            # Put a mutation on every branch.
            t = ts_source.dump_tables()
            j = 0
            for u in range(ts_source.num_nodes - 1):
                for _ in range(2):
                    t.sites.add_row(position=j * d, ancestral_state="0")
                    t.mutations.add_row(site=j, node=u, derived_state="1")
                    j += 1
            ts_source = msprime.load_tables(**t.asdict())
            ts_inferred = infer_from_simulation(ts_source)
            assert ts_inferred.num_trees == 1

def check_single_tree_many_mutations_per_branch():
    num_samples = 10
    for k in range(2, 20):
        print("k = ", k)
        num_mutations = 2 * num_samples * k
        d = 1 / num_mutations
        for seed in range(1, 10):
            ts_source = msprime.simulate(num_samples, length=num_mutations, random_seed=seed)
            # Put a mutation on every branch.
            t = ts_source.dump_tables()
            j = 0
            for u in range(ts_source.num_nodes - 1):
                for _ in range(k):
                    t.sites.add_row(position=j * d, ancestral_state="0")
                    t.mutations.add_row(site=j, node=u, derived_state="1")
                    j += 1
            ts_source = msprime.load_tables(**t.asdict())
            # print(ts_source.tables)
            ts_inferred = infer_from_simulation(ts_source)
            assert ts_inferred.num_trees == 1


def check_single_tree_high_mutation_rate():
    # import daiquiri
    # daiquiri.setup(level="DEBUG")
    # TODO change this test so that we throw a Poisson number of mutations on
    # each branch. It's very hard to have mutations on _every_ branch for larger
    # tree sizes.
    for num_samples in [3, 10, 50]:
        for seed in range(1, 10):
            ts_source = msprime.simulate(num_samples, random_seed=seed, mutation_rate=5000)
            print("sim = ", num_samples, ts_source.num_sites, seed)
            nodes = set()
            for site in ts_source.sites():
                for mutation in site.mutations:
                    nodes.add(mutation.node)
            # if nodes != set(range(ts_source.num_nodes - 1)):
            #     continue
            assert nodes == set(range(ts_source.num_nodes - 1))
            ts_inferred = infer_from_simulation(ts_source)
            # for t in ts_source.trees():
            #     print(t.draw(format="unicode"))
            # print(ts_inferred.num_trees)
            # for t in ts_inferred.trees():
            #     print(t.draw(format="unicode"))
            assert ts_inferred.num_trees == 1


##############################
# Updated code to work with the CLI
##############################


def run_infer(ts, engine="C", path_compression=True, exact_ancestors=False):
    """
    Runs the perfect inference process on the specified tree sequence.
    """

    sample_data = tsinfer.SampleData.from_tree_sequence(ts)

    if exact_ancestors:
        ancestor_data = tsinfer.AncestorData(sample_data)
        tsinfer.build_simulated_ancestors(sample_data, ancestor_data, ts)
        ancestor_data.finalise()
    else:
        ancestor_data = tsinfer.generate_ancestors(sample_data)

    ancestors_ts = tsinfer.match_ancestors(
        sample_data, ancestor_data, path_compression=path_compression,
        engine=engine)
    inferred_ts = tsinfer.match_samples(
        sample_data, ancestors_ts, path_compression=path_compression,
        engine=engine)
    return inferred_ts

def edges_performance_worker(args):
    simulation_args, tree_metrics = args
    before = time.perf_counter()
    smc_ts = msprime.simulate(**simulation_args)
    sim_time = time.perf_counter() - before

    tmp_ts = tsinfer.strip_singletons(smc_ts)
    if tmp_ts.num_sites == 0:
        warnings.warn("Dropping simulation with no variants")
        return {}

    before = time.perf_counter()
    estimated_ancestors_ts = run_infer(smc_ts, exact_ancestors=False)
    estimated_ancestors_time = time.perf_counter() - before
    num_children = []
    for edgeset in estimated_ancestors_ts.edgesets():
        num_children.append(len(edgeset.children))
    estimated_ancestors_num_children = np.array(num_children)

    before = time.perf_counter()
    exact_ancestors_ts = run_infer(smc_ts, exact_ancestors=True)
    exact_ancestors_time = time.perf_counter() - before
    num_children = []
    for edgeset in exact_ancestors_ts.edgesets():
        num_children.append(len(edgeset.children))
    exact_ancestors_num_children = np.array(num_children)

    results = {
        "sim_time": sim_time,
        "estimated_anc_time": estimated_ancestors_time,
        "exact_anc_time": exact_ancestors_time,
        "num_sites": smc_ts.num_sites,
        "source_num_trees": smc_ts.num_trees,
        "estimated_anc_trees": estimated_ancestors_ts.num_trees,
        "exact_anc_trees": exact_ancestors_ts.num_trees,
        "source_edges": smc_ts.num_edges,
        "estimated_anc_edges": estimated_ancestors_ts.num_edges,
        "exact_anc_edges": exact_ancestors_ts.num_edges,
        "estimated_anc_max_children": np.max(estimated_ancestors_num_children),
        "estimated_anc_mean_children":
            np.mean(estimated_ancestors_num_children),
        "exact_anc_max_children": np.max(exact_ancestors_num_children),
        "exact_anc_mean_children":
            np.mean(exact_ancestors_num_children),
    }
    results.update(simulation_args)
    if tree_metrics:
        before = time.perf_counter()
        breakpoints, kc_distance = tsinfer.compare(smc_ts, exact_ancestors_ts)
        d = breakpoints[1:] - breakpoints[:-1]
        d /= breakpoints[-1]
        exact_anc_kc_distance_weighted = np.sum(kc_distance * d)
        exact_anc_perfect_trees = np.sum((kc_distance == 0) * d)
        exact_anc_kc_mean = np.mean(kc_distance)
        breakpoints, kc_distance = tsinfer.compare(smc_ts, estimated_ancestors_ts)
        d = breakpoints[1:] - breakpoints[:-1]
        d /= breakpoints[-1]
        estimated_anc_kc_distance_weighted = np.sum(kc_distance * d)
        estimated_anc_perfect_trees = np.sum((kc_distance == 0) * d)
        estimated_anc_kc_mean = np.mean(kc_distance)
        tree_metrics_time = time.perf_counter() - before
        results.update({
            "tree_metrics_time": tree_metrics_time,
            "exact_anc_kc_distance_weighted": exact_anc_kc_distance_weighted,
            "exact_anc_perfect_trees": exact_anc_perfect_trees,
            "exact_anc_kc_mean": exact_anc_kc_mean,
            "estimated_anc_kc_distance_weighted": estimated_anc_kc_distance_weighted,
            "estimated_anc_perfect_trees": estimated_anc_perfect_trees,
            "estimated_anc_kc_mean": estimated_anc_kc_mean,
        })
    return results


def run_edges_performance(args):
    num_lengths = 10
    MB = 10**6

    work = []
    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)
    for L in np.linspace(0, args.length, num_lengths + 1)[1:]:
        for _ in range(args.num_replicates):
            sim_args = {
                "sample_size": args.sample_size,
                "length": L * MB,
                "recombination_rate": args.recombination_rate,
                "mutation_rate": args.mutation_rate,
                "Ne": 10**4,
                "model": "smc_prime",
                "random_seed": rng.randint(1, 2**30)}
            work.append((sim_args, args.compute_tree_metrics))

    random.shuffle(work)
    progress = tqdm.tqdm(total=len(work), disable=not args.progress)
    results = []
    try:
        with concurrent.futures.ProcessPoolExecutor(args.num_processes) as executor:
            for result in executor.map(edges_performance_worker, work):
                results.append(result)
                progress.update()

    except KeyboardInterrupt:
        pass
    progress.close()

    df = pd.DataFrame(results)
    df.length /= MB
    dfg = df.groupby(df.length).mean()
    # print(dfg.estimated_anc_edges.describe())
    print(dfg)

    name_format = os.path.join(
        args.destination_dir,
        "ancestors_n={}_L={}_mu={}_rho={}_{{}}.png".format(
            args.sample_size, args.length, args.mutation_rate, args.recombination_rate))

    plt.plot(
        dfg.num_sites, dfg.estimated_anc_edges / dfg.source_edges,
        label="estimated ancestors")
    plt.plot(
        dfg.num_sites, dfg.exact_anc_edges / dfg.source_edges,
        label="exact ancestors")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    plt.ylabel("inferred # edges / source # edges")
    plt.xlabel("Num sites")
    plt.legend()
    plt.savefig(name_format.format("edges"))
    plt.clf()

    plt.plot(
        dfg.num_sites, dfg.estimated_anc_mean_children,
        label="estimated ancestors mean", color="blue")
    plt.plot(
        dfg.num_sites, dfg.estimated_anc_max_children,
        label="estimated ancestors max", color="blue", linestyle=":")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    plt.plot(
        dfg.num_sites, dfg.exact_anc_mean_children,
        label="exact ancestors mean", color="red")
    plt.plot(
        dfg.num_sites, dfg.exact_anc_max_children,
        label="exact ancestors max", color="red", linestyle=":")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    plt.ylabel("num_children")
    plt.xlabel("Num sites")
    plt.legend()
    plt.savefig(name_format.format("num_children"))
    plt.clf()



    if args.compute_tree_metrics:
        plt.plot(
            dfg.num_sites, dfg.estimated_anc_kc_distance_weighted,
            label="estimated ancestors")
        plt.plot(
            dfg.num_sites, dfg.exact_anc_kc_distance_weighted,
            label="exact ancestors")
        plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
            args.sample_size, args.mutation_rate, args.recombination_rate,
            args.num_replicates))
        plt.ylabel("Distance weighted KC metric")
        plt.xlabel("Num sites")
        plt.legend()
        plt.savefig(name_format.format("kc_distance_weighted"))
        plt.clf()

        plt.plot(
            dfg.num_sites, dfg.estimated_anc_kc_mean,
            label="estimated ancestors")
        plt.plot(
            dfg.num_sites, dfg.exact_anc_kc_mean,
            label="exact ancestors")
        plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
            args.sample_size, args.mutation_rate, args.recombination_rate,
            args.num_replicates))
        plt.ylabel("Mean KC metric")
        plt.xlabel("Num sites")
        plt.legend()
        plt.savefig(name_format.format("kc_mean"))
        plt.clf()

        plt.plot(
            dfg.num_sites, dfg.estimated_anc_perfect_trees,
            label="estimated ancestors")
        plt.plot(
            dfg.num_sites, dfg.exact_anc_perfect_trees,
            label="exact ancestors")
        plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
            args.sample_size, args.mutation_rate, args.recombination_rate,
            args.num_replicates))
        plt.ylabel("Mean KC metric")
        plt.xlabel("Num sites")
        plt.legend()
        plt.savefig(name_format.format("perfect_trees"))
        plt.clf()


def get_num_sample_edges(ts):
    """
    Returns the number of edges where the child is a sample.
    """
    count = 0
    for e in ts.edges():
        node = ts.node(e.child)
        count += node.is_sample()
    return count


def effective_recombination_worker(args):
    simulation_args = args
    smc_ts = msprime.simulate(**simulation_args)
    estimated_ancestors_ts = run_infer(smc_ts, exact_ancestors=False)
    exact_ancestors_ts = run_infer(smc_ts, exact_ancestors=True)
    results = {
        "num_sites": smc_ts.num_sites,
        "source_num_trees": smc_ts.num_trees,
        "source_sample_edges": get_num_sample_edges(smc_ts),
        "estimated_anc_sample_edges": get_num_sample_edges(estimated_ancestors_ts),
        "exact_anc_sample_edges": get_num_sample_edges(exact_ancestors_ts),
    }
    results.update(simulation_args)
    return results


def run_effective_recombination(args):
    num_sample_sizes = 10
    MB = 10**6

    work = []
    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)
    for n in np.linspace(0, args.sample_size, num_sample_sizes + 1)[1:].astype(int):
        for _ in range(args.num_replicates):
            sim_args = {
                "sample_size": n,
                "length": args.length * MB,
                "recombination_rate": args.recombination_rate,
                "mutation_rate": args.mutation_rate,
                "Ne": 10**4,
                "model": "smc_prime",
                "random_seed": rng.randint(1, 2**30)}
            work.append(sim_args)

    random.shuffle(work)
    progress = tqdm.tqdm(total=len(work), disable=not args.progress)
    results = []
    try:
        with concurrent.futures.ProcessPoolExecutor(args.num_processes) as executor:
            for result in executor.map(effective_recombination_worker, work):
                results.append(result)
                progress.update()

    except KeyboardInterrupt:
        pass
    progress.close()

    df = pd.DataFrame(results)
    df.length /= MB
    df.source_sample_edges -= df.sample_size
    df.estimated_anc_sample_edges -= df.sample_size
    df.exact_anc_sample_edges -= df.sample_size

    dfg = df.groupby(df.sample_size).mean()
    print(dfg)

    name_format = os.path.join(
        args.destination_dir,
        "effective_recombination_n={}_L={}_mu={}_rho={}_{{}}.png".format(
            args.sample_size, args.length, args.mutation_rate, args.recombination_rate))

    plt.plot(dfg.source_sample_edges, label="source")
    plt.plot(dfg.estimated_anc_sample_edges, label="estimated ancestors")
    plt.plot(dfg.exact_anc_sample_edges, label="exact ancestors")
    plt.ylabel("Number of sample edges - n")
    plt.xlabel("Sample size")
    plt.legend()
    plt.savefig(name_format.format("sample_edges"))
    plt.clf()


def unrank(samples, n):
    """
    Unranks the specified set of samples from a possible n into its position
    in a lexicographically sorted list of bitstrings.
    """
    bitstring = np.zeros(n, dtype=int)
    for s in samples:
        bitstring[s] = 1
    mult = 2**np.arange(n, dtype=int)
    unranked = np.sum(mult * bitstring)
    return unranked


def edge_plot(ts, filename):
    n = ts.num_samples
    pallete = sns.color_palette("husl", 2**n - 1)
    lines = []
    colours = []
    for tree in ts.trees():
        left, right = tree.interval
        for u in tree.nodes():
            for c in tree.children(u):
                lines.append([(left, c), (right, c)])
                colours.append(pallete[unrank(tree.samples(c), n)])

    lc = mc.LineCollection(lines, linewidths=2, colors=colours)
    fig, ax = plt.subplots()
    ax.add_collection(lc)
    ax.autoscale()
    plt.savefig(filename)
    plt.clf()



def run_hotspot_analysis(args):
    MB = 10**6
    L = args.length * MB

    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)

    breakpoints = np.linspace(0, L, args.num_hotspots + 2)
    end = breakpoints[1:-1] + L * args.hotspot_width
    breakpoints = np.hstack([breakpoints, end])
    breakpoints.sort()
    rates = np.zeros_like(breakpoints)
    rates[:-1] = args.recombination_rate
    # Set the odd elements of the array to be hotspots.
    rates[1::2] *= args.hotspot_intensity
    recomb_map = msprime.RecombinationMap(list(breakpoints), list(rates))

    sim_args = {
        "sample_size": args.sample_size,
        "recombination_map": recomb_map,
        "mutation_rate": args.mutation_rate,
        "Ne": 10**4,
        "random_seed": rng.randint(1, 2**30)}
    ts = msprime.simulate(**sim_args)
    print("simulated ", ts.num_trees, "trees and", ts.num_sites, "sites")

    inferred_ts = run_infer(ts)

    num_bins = 100
    hotspot_breakpoints = breakpoints

    for density in [True, False]:
        for x in hotspot_breakpoints[1:-1]:
            plt.axvline(x=x, color="k", ls=":")
        breakpoints = np.array(list(inferred_ts.breakpoints()))
        v, bin_edges = np.histogram(breakpoints, num_bins, density=density)
        plt.plot(bin_edges[:-1], v, label="inferred")
        breakpoints = np.array(list(ts.breakpoints()))
        v, bin_edges = np.histogram(breakpoints, num_bins, density=density)
        plt.plot(bin_edges[:-1], v, label="source")
        plt.ylabel("Number of breakpoints")
        plt.legend()

        name_format = os.path.join(
            args.destination_dir,
            "hotspots_n={}_L={}_mu={}_rho={}_N={}_I={}_W={}_{{}}.png".format(
                args.sample_size, args.length, args.mutation_rate, args.recombination_rate,
                args.num_hotspots, args.hotspot_intensity, args.hotspot_width))
        plt.savefig(name_format.format("breakpoints_density={}".format(density)))
        plt.clf()

    print("Generating edge plots")
    # TODO add option for colour mapping.
    edge_plot(ts, name_format.format("source_edges"))
    edge_plot(inferred_ts, name_format.format("dest_edges"))


def error_analysis_worker(args):
    simulation_args, input_error, inference_error = args
    before = time.perf_counter()
    ts = msprime.simulate(**simulation_args)

    # V = generate_samples(ts, input_error)
    ts = tsinfer.insert_errors(ts, inference_error)
    V = ts.genotype_matrix()
    inferred_ts = tsinfer.infer(
        V, [site.position for site in ts.sites()], sample_error=0,
        sequence_length=ts.sequence_length,
        recombination_rate=simulation_args["recombination_rate"])

    results = {
        "input_error": input_error,
        "inference_error": inference_error,
        "num_sites": ts.num_sites,
        "source_num_trees": ts.num_trees,
        "inferred_num_trees": inferred_ts.num_trees,
        "source_edges": ts.num_edges,
        "inferred_edges": inferred_ts.num_edges,
    }
    results.update(simulation_args)

    breakpoints, kc_distance = tsinfer.compare(ts, inferred_ts)
    d = breakpoints[1:] - breakpoints[:-1]
    d /= breakpoints[-1]
    kc_distance_weighted = np.sum(kc_distance * d)
    kc_mean = np.mean(kc_distance)
    results.update({
        "kc_distance_weighted": kc_distance_weighted,
        "kc_mean": kc_mean,
    })
    return results


def run_error_analysis(args):
    MB = 10**6
    L = args.length * MB

    work = []
    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)
    inference_errors = [
        0, args.error_probability / 10, args.error_probability,
        args.error_probability * 10]
    for e in inference_errors:
        for _ in range(args.num_replicates):
            sim_args = {
                "sample_size": args.sample_size,
                "length": L,
                "recombination_rate": args.recombination_rate,
                "mutation_rate": args.mutation_rate,
                "Ne": 10**4,
                "random_seed": rng.randint(1, 2**30)}
            work.append((sim_args, args.error_probability, e))

    random.shuffle(work)
    progress = tqdm.tqdm(total=len(work), disable=not args.progress)
    results = []
    try:
        with concurrent.futures.ProcessPoolExecutor(args.num_processes) as executor:
            for result in executor.map(error_analysis_worker, work):
                results.append(result)
                progress.update()

    except KeyboardInterrupt:
        pass
    progress.close()

    df = pd.DataFrame(results)

    name_format = os.path.join(
        args.destination_dir,
        "error_n={}_L={}_mu={}_rho={}_e={}_{{}}.png".format(
            args.sample_size, args.length, args.mutation_rate,
            args.recombination_rate, args.error_probability))

    sns.boxplot(x="inference_error", y="kc_mean", data=df)
    plt.ylabel("KC distance")
    plt.xlabel("Error parameter")
    plt.savefig(name_format.format("kc"))
    plt.clf()

    # plt.plot(dfg.inferred_edges / dfg.source_edges)
    # plt.axvline(args.error_probability, ls="--")
    # plt.ylabel("inferred # edges / source # edges")
    # plt.xlabel("Error parameter")
    # plt.savefig(name_format.format("edges"))
    # plt.clf()

    # plt.plot(dfg.kc_mean)
    # plt.axvline(args.error_probability, ls="--")
    # plt.ylabel("KC distance")
    # plt.xlabel("Error parameter")
    # plt.savefig(name_format.format("kc"))
    # plt.clf()



def ancestor_properties_worker(args):
    simulation_args, compute_exact = args
    ts = msprime.simulate(**simulation_args)

    sample_data = tsinfer.SampleData.from_tree_sequence(ts)
    estimated_anc = tsinfer.generate_ancestors(sample_data)
    estimated_anc_length = estimated_anc.end[:] - estimated_anc.start[:]
    focal_sites = estimated_anc.focal_sites[:]
    estimated_anc_focal_distance = np.zeros(estimated_anc.num_ancestors)
    for j in range(estimated_anc.num_ancestors):
        focal = focal_sites[j]
        if len(focal) > 0:
            estimated_anc_focal_distance[j] = focal[-1] - focal[0]

    results = {
        "num_sites": ts.num_sites,
        "num_trees": ts.num_trees,
        "estimated_anc_num": estimated_anc.num_ancestors,
        "estimated_anc_mean_len": np.mean(estimated_anc_length),
        "estimated_anc_mean_focal_distance": np.mean(estimated_anc_focal_distance),
    }

    if compute_exact:
        exact_anc = tsinfer.AncestorData(sample_data)
        tsinfer.build_simulated_ancestors(sample_data, exact_anc, ts)
        exact_anc.finalise()
        exact_anc_length = exact_anc.end[:] - exact_anc.start[:]

        focal_sites = exact_anc.focal_sites[:]
        exact_anc_focal_distance = np.zeros(exact_anc.num_ancestors)
        for j in range(exact_anc.num_ancestors):
            focal = focal_sites[j]
            if len(focal) > 0:
                exact_anc_focal_distance[j] = focal[-1] - focal[0]
        results.update({
            "exact_anc_num": exact_anc.num_ancestors,
            "exact_anc_mean_len": np.mean(exact_anc_length),
            "exact_anc_mean_focal_distance": np.mean(exact_anc_focal_distance),
        })

    results.update(simulation_args)
    return results


def run_ancestor_properties(args):
    num_lengths = 10
    MB = 10**6

    work = []
    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)
    for L in np.linspace(0, args.length, num_lengths + 1)[1:]:
        for _ in range(args.num_replicates):
            sim_args = {
                "sample_size": args.sample_size,
                "length": L * MB,
                "recombination_rate": args.recombination_rate,
                "mutation_rate": args.mutation_rate,
                "Ne": 10**4,
                "model": "smc_prime",
                "random_seed": rng.randint(1, 2**30)}
            work.append((sim_args, not args.skip_exact))

    random.shuffle(work)
    progress = tqdm.tqdm(total=len(work), disable=not args.progress)
    results = []
    try:
        with concurrent.futures.ProcessPoolExecutor(args.num_processes) as executor:
            for result in executor.map(ancestor_properties_worker, work):
                results.append(result)
                progress.update()

    except KeyboardInterrupt:
        pass
    progress.close()

    df = pd.DataFrame(results)
    df.length /= MB
    dfg = df.groupby(df.length).mean()
    print(dfg)

    name_format = os.path.join(
        args.destination_dir, "anc-prop_n={}_L={}_mu={}_rho={}_{{}}.png".format(
        args.sample_size, args.length, args.mutation_rate, args.recombination_rate))

    # plt.plot(
    #     dfg.num_sites, dfg.exact_anc_num / dfg.estimated_anc_num,
    #     label="")
        # label="exact ancestors")
    plt.plot(dfg.num_sites, dfg.estimated_anc_num, label="estimated ancestors")
    if not args.skip_exact:
        plt.plot(dfg.num_sites, dfg.exact_anc_num, label="exact ancestors")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    # plt.ylabel("inferred # ancestors / exact # ancestors")
    plt.xlabel("Num sites")
    plt.legend()
    plt.savefig(name_format.format("num"))
    plt.clf()

    plt.plot(dfg.num_sites, dfg.estimated_anc_mean_len, label="estimated ancestors")
    if not args.skip_exact:
        plt.plot(dfg.num_sites, dfg.exact_anc_mean_len, label="exact ancestors")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    # plt.ylabel("inferred # ancestors / exact # ancestors")
    plt.xlabel("Num sites")
    plt.legend()
    plt.savefig(name_format.format("mean_len"))
    plt.clf()

    plt.plot(
        dfg.num_sites, dfg.estimated_anc_mean_focal_distance,
        label="estimated ancestors")
    if not args.skip_exact:
        plt.plot(
            dfg.num_sites, dfg.exact_anc_mean_focal_distance,
            label="exact ancestors")
    plt.title("n = {}, mut_rate={}, rec_rate={}, reps={}".format(
        args.sample_size, args.mutation_rate, args.recombination_rate,
        args.num_replicates))
    # plt.ylabel("inferred # ancestors / exact # ancestors")
    plt.xlabel("Num sites")
    plt.legend()
    plt.savefig(name_format.format("mean_focal_distance"))
    plt.clf()

def running_mean(x, N):
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)

def run_ancestor_comparison(args):
    MB = 10**6
    rng = random.Random(args.random_seed)
    sim_args = {
        "sample_size": args.sample_size,
        "length": args.length * MB,
        "recombination_rate": args.recombination_rate,
        "mutation_rate": args.mutation_rate,
        "Ne": 10**4,
        "model": "smc_prime",
        "random_seed": rng.randint(1, 2**30)}
    ts = msprime.simulate(**sim_args)

    # ts  = tsinfer.insert_errors(ts, args.error_probability, seed=args.random_seed)
    V = generate_samples(ts, args.error_probability)

    sample_data = tsinfer.SampleData(sequence_length=ts.sequence_length)
    for j, v in enumerate(V):
        sample_data.add_site(j, ["0", "1"], v)
    sample_data.finalise()

    estimated_anc = tsinfer.generate_ancestors(sample_data)
    estimated_anc_length = estimated_anc.end[1:] - estimated_anc.start[1:]

    exact_anc = tsinfer.AncestorData(sample_data)
    tsinfer.build_simulated_ancestors(sample_data, exact_anc, ts)
    exact_anc.finalise()
    exact_anc_length = exact_anc.end[1:] - exact_anc.start[1:]


    name_format = os.path.join(
        args.destination_dir, "anc-comp_n={}_L={}_mu={}_rho={}_err={}_{{}}".format(
        args.sample_size, args.length, args.mutation_rate, args.recombination_rate,
        args.error_probability))
    if args.store_data:
        filename = name_format.format("length.json")
        data = {
            "exact_ancestors": exact_anc_length.tolist(),
            "estimated_ancestors": estimated_anc_length.tolist()}
        with open(filename, "w") as f:
            json.dump(data, f)

    plt.hist([exact_anc_length, estimated_anc_length], label=["Exact", "Estimated"])
    plt.legend()
    plt.savefig(name_format.format("length-dist.png"))
    plt.clf()

    frequency = estimated_anc.time[:][1:]
    print(estimated_anc_length[frequency==2])
    plt.hist(estimated_anc_length[frequency==2], bins=50)
    plt.xlabel("doubleton ancestor length")
    plt.savefig(name_format.format("doubleton-length-dist.png"))
    plt.clf()

    # Because we have different numbers of ancestors, we need to rescale time
    # somehow to display them on the same axis. We just map time linearly into
    # 0,1. Possibly this is misleading.
    nbins = 100
    x = running_mean(exact_anc_length, nbins)
    # x = exact_anc_length
    plt.plot(np.linspace(0, 1, x.shape[0]), x, label="Exact")
    x = running_mean(estimated_anc_length, nbins)
    # x = estimated_anc_length
    plt.plot(np.linspace(0, 1, x.shape[0]), x, label="Estimated")
    plt.xlabel("Time (oldest to youngest)")
    plt.ylabel("Length")
    plt.legend()
    plt.savefig(name_format.format("length-time.png"))
    plt.clf()

def get_node_degree_by_depth(ts):
    """
    Returns a tuple (degree, depth) for each node in each tree in the
    specified tree sequence.
    """
    degree = []
    depth = []
    for tree in ts.trees():
        stack = [(tree.root, 0)]
        while len(stack) > 0:
            u, d = stack.pop()
            if len(tree.children(u)) > 0:
                degree.append(len(tree.children(u)))
                depth.append(d)
            for v in tree.children(u):
                stack.append((v, d + 1))
    return np.array(degree), np.array(depth)


def run_node_degree(args):
    MB = 10**6
    rng = random.Random()
    if args.random_seed is not None:
        rng.seed(args.random_seed)
    sim_args = {
        "sample_size": args.sample_size,
        "length": args.length * MB,
        "recombination_rate": args.recombination_rate,
        "mutation_rate": args.mutation_rate,
        "Ne": 10**4,
        "model": "smc_prime",
        "random_seed": rng.randint(1, 2**30)}
    smc_ts = msprime.simulate(**sim_args)

    engine= "C"
    df = pd.DataFrame()
    for path_compression in [True, False]:
        estimated_ancestors_ts = run_infer(
            smc_ts, engine=engine, exact_ancestors=False, path_compression=path_compression)
        degree, depth = get_node_degree_by_depth(estimated_ancestors_ts)
        df = df.append(pd.DataFrame({
            "degree": degree, "depth": depth, "type":"estimated",
            "path_compression": path_compression}))
        exact_ancestors_ts = run_infer(
            smc_ts, engine=engine, exact_ancestors=True, path_compression=path_compression)
        degree, depth = get_node_degree_by_depth(exact_ancestors_ts)
        df = df.append(pd.DataFrame({
            "degree": degree, "depth": depth, "type":"exact",
            "path_compression": path_compression}))

    name_format = os.path.join(
        args.destination_dir, "node-degree_n={}_L={}_mu={}_rho={}_{{}}.png".format(
        args.sample_size, args.length, args.mutation_rate, args.recombination_rate))
    print(df.describe())

    sns.factorplot(x="depth", y="degree",
            hue="path_compression", col="type",
            data=df, kind="bar")
    plt.savefig(name_format.format("path-compression"))
    plt.clf()

    sns.barplot(x="depth", y="degree", hue="type", data=df[df.path_compression])
    plt.savefig(name_format.format("length"))
    plt.clf()


def multiple_recombinations(ts):
    """
    Returns true if the specified tree sequence contains multiple recombinations.
    """
    for _, e_out, _ in ts.edge_diffs():
        if len(e_out) > 4:
            return True
    return False

def run_perfect_inference(args):
    for seed in range(1, args.num_replicates + 1):
        base_ts = msprime.simulate(
            args.sample_size, Ne=10**4, length=args.length * 10**6,
            recombination_rate=1e-8, random_seed=seed, model="smc_prime")
        print("simulated ts with n={} and {} trees; seed={}".format(
            base_ts.num_samples, base_ts.num_trees, seed))
        if multiple_recombinations(base_ts):
            print("Multiple recombinations; skipping")
            continue
        ts, inferred_ts = tsinfer.run_perfect_inference(
            base_ts, num_threads=args.num_threads,
            engine=args.engine, extended_checks=args.extended_checks,
            time_chunking=not args.no_time_chunking,
            path_compression=args.path_compression)
        print("n={} num_trees={} num_sites={}".format(
            ts.num_samples, ts.num_trees, ts.num_sites))
        assert ts.num_samples == inferred_ts.num_samples
        assert ts.num_sites == inferred_ts.num_sites
        if args.path_compression:
            _, distances = tsinfer.compare(ts, inferred_ts)
            assert np.all(distances == 0)
        else:
            assert ts.tables.edges == inferred_ts.tables.edges
            assert np.all(ts.tables.sites.position == inferred_ts.tables.sites.position)
            assert ts.tables.mutations == inferred_ts.tables.mutations
            assert np.array_equal(ts.tables.nodes.flags, inferred_ts.tables.nodes.flags)
            assert np.any(ts.tables.nodes.time != inferred_ts.tables.nodes.time)


def setup_logging(args):
    log_level = "WARN"
    if args.verbosity > 0:
        log_level = "INFO"
    if args.verbosity > 1:
        log_level = "DEBUG"
    if args.log_section is None:
        daiquiri.setup(level=log_level)
    else:
        daiquiri.setup(level="WARN")
        logger = logging.getLogger(args.log_section)
        logger.setLevel(log_level)


if __name__ == "__main__":

    top_parser = argparse.ArgumentParser(
        description="Simple inferface for running various tsinfer evaluations.")
    top_parser.add_argument(
        "-V", "--version", action='version',
        version='%(prog)s {}'.format(tsinfer.__version__))

    subparsers = top_parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    parser = subparsers.add_parser(
        "perfect-inference", aliases=["pi"],
        help="Runs the perfect inference process on simulated tree sequences.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_perfect_inference)
    parser.add_argument("--engine", default="C")
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument("--num-replicates", "-R", type=int, default=1)
    parser.add_argument("--num-threads", "-t", type=int, default=0)
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")
    parser.add_argument(
        "--extended-checks", "-X", action="store_true",
        help="Enable extra consistency checking (slow)")
    parser.add_argument(
        "--no-time-chunking", action="store_true",
        help="Disable time-chunking to give each ancestor a distinct time.")
    parser.add_argument(
        "--path-compression", "-c", action="store_true",
        help="Turn on path compression. Makes verification much slower.")

    parser = subparsers.add_parser(
        "edges-performance", aliases=["ep"],
        help="Runs a plot showing performance in terms of the edge ratio.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_edges_performance)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument("--num-replicates", "-R", type=int, default=10)
    parser.add_argument("--num-processes", "-P", type=int, default=None)
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--compute-tree-metrics", "-T", action="store_true",
        help="Compute tree metrics")
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")

    parser = subparsers.add_parser(
        "effective-recombination", aliases=["er"],
        help="Shows the effective recombination rate against sample size.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_effective_recombination)
    parser.add_argument("--sample-size", "-n", type=int, default=1000)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument("--num-replicates", "-R", type=int, default=10)
    parser.add_argument("--num-processes", "-P", type=int, default=None)
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")

    parser = subparsers.add_parser(
        "hotspot-analysis", aliases=["ha"],
        help="Runs plots analysing the effects of recombination hotspots.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_hotspot_analysis)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument(
        "--num-hotspots", "-N", type=int, default=1,
        help="Number of hotspots")
    parser.add_argument(
        "--hotspot-intensity", "-I", type=float, default=10,
        help="Intensity of hotspots relative to background.")
    parser.add_argument(
        "--hotspot-width", "-W", type=float, default=0.01,
        help="Width of hotspots as a fraction of total genome length.")
    parser.add_argument("--num-replicates", "-R", type=int, default=10)
    parser.add_argument("--num-processes", "-P", type=int, default=None)
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")

    parser = subparsers.add_parser(
        "error-analysis", aliases=["ea"],
        help="Runs plots analysing the effects of inserted errors.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_error_analysis)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument(
        "--error-probability", "-e", type=float, default=0.01,
        help="Probability of errors.")
    parser.add_argument("--num-replicates", "-R", type=int, default=10)
    parser.add_argument("--num-processes", "-P", type=int, default=None)
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")

    parser = subparsers.add_parser(
        "ancestor-properties", aliases=["ap"],
        help="Runs plots showing the properties of estimated ancestors.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_ancestor_properties)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument("--num-replicates", "-R", type=int, default=10)
    parser.add_argument("--num-processes", "-P", type=int, default=None)
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")
    parser.add_argument(
        "--skip-exact", "-S", action="store_true",
        help="Skip computing the exact ancestors")

    parser = subparsers.add_parser(
        "ancestor-comparison", aliases=["ac"],
        help="Runs plots comparing the real and simulated ancestors for a single instance.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_ancestor_comparison)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument(
        "--error-probability", "-e", type=float, default=0,
        help="Error probability")
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")
    parser.add_argument(
        "--store-data", "-S", action="store_true",
        help="Store the raw data.")

    parser = subparsers.add_parser(
        "node-degree", aliases=["nd"],
        help="Plots node degree vs depth in the tree.")
    cli.add_logging_arguments(parser)
    parser.set_defaults(runner=run_node_degree)
    parser.add_argument("--sample-size", "-n", type=int, default=10)
    parser.add_argument(
        "--length", "-l", type=float, default=1, help="Sequence length in MB")
    parser.add_argument(
        "--recombination-rate", "-r", type=float, default=1e-8,
        help="Recombination rate")
    parser.add_argument(
        "--mutation-rate", "-u", type=float, default=1e-8,
        help="Mutation rate")
    parser.add_argument("--random-seed", "-s", type=int, default=None)
    parser.add_argument("--destination-dir", "-d", default="")

    args = top_parser.parse_args()
    cli.setup_logging(args)
    args.runner(args)



    # check_basic_performance()
    # check_effect_error_param(float(sys.argv[1]))
    # plot_basic_performance()
    # check_variable_recomb()
    # check_single_tree_one_mutation_per_branch()
    # check_single_tree_two_mutations_per_branch()
    # check_single_tree_many_mutations_per_branch()
    # check_single_tree_high_mutation_rate()
    # check_many_trees_one_mutation_per_branch()





#
# Copyright (C) 2018 University of Oxford
#
# This file is part of tsinfer.
#
# tsinfer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tsinfer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tsinfer.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for the data files.
"""

import unittest
import tempfile
import os.path
import datetime
import warnings

import numpy as np
import msprime
import numcodecs.blosc as blosc
import zarr

import tsinfer
import tsinfer.formats as formats
import tsinfer.exceptions as exceptions


class DataContainerMixin(object):
    """
    Common tests for the the data container classes."
    """
    def test_load(self):
        self.assertRaises(IsADirectoryError, self.tested_class.load, "/")
        self.assertRaises(
            FileNotFoundError, self.tested_class.load, "/file/does/not/exist")
        bad_format_files = ["LICENSE", "/dev/urandom"]
        for bad_format_file in bad_format_files:
            self.assertTrue(os.path.exists(bad_format_file))
            self.assertRaises(
                exceptions.FileFormatError, formats.SampleData.load, bad_format_file)


class TestSampleData(unittest.TestCase, DataContainerMixin):
    """
    Test cases for the sample data file format.
    """
    tested_class = formats.SampleData

    def get_example_ts(self, sample_size, sequence_length):
        return msprime.simulate(
            sample_size, recombination_rate=1, mutation_rate=10,
            length=sequence_length, random_seed=100)

    def verify_data_round_trip(self, ts, input_file):
        self.assertGreater(ts.num_sites, 1)
        for v in ts.variants():
            input_file.add_site(v.site.position, v.genotypes, v.alleles)
        input_file.finalise()
        self.assertEqual(input_file.format_version, formats.SampleData.FORMAT_VERSION)
        self.assertEqual(input_file.format_name, formats.SampleData.FORMAT_NAME)
        self.assertEqual(input_file.num_samples, ts.num_samples)
        self.assertEqual(input_file.sequence_length, ts.sequence_length)
        self.assertEqual(input_file.num_sites, ts.num_sites)
        self.assertEqual(input_file.sites_genotypes.dtype, np.uint8)
        self.assertEqual(input_file.sites_position.dtype, np.float64)
        # Take copies to avoid decompressing the data repeatedly.
        genotypes = input_file.sites_genotypes[:]
        position = input_file.sites_position[:]
        alleles = input_file.sites_alleles[:]
        for j, variant in enumerate(ts.variants()):
            self.assertEqual(variant.site.position, position[j])
            self.assertTrue(np.all(variant.genotypes == genotypes[j]))
            self.assertEqual(alleles[j], list(variant.alleles))

    def test_defaults_with_path(self):
        ts = self.get_example_ts(10, 10)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "samples.tmp")
            input_file = formats.SampleData(
                path=filename, sequence_length=ts.sequence_length)
            self.verify_data_round_trip(ts, input_file)
            compressor = formats.DEFAULT_COMPRESSOR
            for _, array in input_file.arrays():
                self.assertEqual(array.compressor, compressor)
            with tsinfer.load(filename) as other:
                self.assertEqual(other, input_file)

    def test_defaults_no_path(self):
        ts = self.get_example_ts(10, 10)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)
        for _, array in input_file.arrays():
            self.assertEqual(array.compressor, None)

    def test_from_tree_sequence(self):
        ts = self.get_example_ts(10, 10)
        sd1 = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, sd1)
        sd2 = formats.SampleData.from_tree_sequence(ts)
        self.assertTrue(sd1.data_equal(sd2))

    def test_chunk_size(self):
        ts = self.get_example_ts(4, 2)
        self.assertGreater(ts.num_sites, 50)
        for chunk_size in [1, 2, 3, ts.num_sites - 1, ts.num_sites, ts.num_sites + 1]:
            input_file = formats.SampleData(
                sequence_length=ts.sequence_length, chunk_size=chunk_size)
            self.verify_data_round_trip(ts, input_file)
            for name, array in input_file.arrays():
                self.assertEqual(array.chunks[0], chunk_size)
                if name.endswith("genotypes"):
                    self.assertEqual(array.chunks[1], chunk_size)

    def test_filename(self):
        ts = self.get_example_ts(14, 15)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "samples.tmp")
            input_file = formats.SampleData(
                sequence_length=ts.sequence_length, path=filename)
            self.assertTrue(os.path.exists(filename))
            self.assertFalse(os.path.isdir(filename))
            self.verify_data_round_trip(ts, input_file)
            other_input_file = formats.SampleData.load(filename)
            self.assertIsNot(other_input_file, input_file)
            self.assertEqual(other_input_file, input_file)

    def test_chunk_size_file_equal(self):
        ts = self.get_example_ts(13, 15)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            files = []
            for chunk_size in [5, 7]:
                filename = os.path.join(tempdir, "samples_{}.tmp".format(chunk_size))
                files.append(filename)
                input_file = formats.SampleData(
                    sequence_length=ts.sequence_length, path=filename,
                    chunk_size=chunk_size)
                self.verify_data_round_trip(ts, input_file)
                self.assertEqual(
                    input_file.sites_genotypes.chunks, (chunk_size, chunk_size))
            # Now reload the files and check they are equal
            input_file0 = formats.SampleData.load(files[0])
            input_file1 = formats.SampleData.load(files[1])
            # Can't use eq here because UUIDs will be equal.
            self.assertTrue(input_file0.data_equal(input_file1))

    def test_compressor(self):
        ts = self.get_example_ts(11, 17)
        compressors = [
           None, formats.DEFAULT_COMPRESSOR,
           blosc.Blosc(cname='zlib', clevel=1, shuffle=blosc.NOSHUFFLE)
        ]
        for compressor in compressors:
            input_file = formats.SampleData(
                sequence_length=ts.sequence_length, compressor=compressor)
            self.verify_data_round_trip(ts, input_file)
            for _, array in input_file.arrays():
                self.assertEqual(array.compressor, compressor)

    def test_multichar_alleles(self):
        ts = self.get_example_ts(5, 17)
        t = ts.dump_tables()
        t.sites.clear()
        t.mutations.clear()
        for site in ts.sites():
            t.sites.add_row(site.position, ancestral_state="A" * (site.id + 1))
            for mutation in site.mutations:
                t.mutations.add_row(
                    site=site.id, node=mutation.node, derived_state="T" * site.id)
        ts = t.tree_sequence()
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)

    def test_str(self):
        ts = self.get_example_ts(5, 3)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)
        self.assertGreater(len(str(input_file)), 0)

    def test_eq(self):
        ts = self.get_example_ts(5, 3)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)
        self.assertTrue(input_file == input_file)
        self.assertFalse(input_file == [])
        self.assertFalse({} == input_file)

    def test_provenance(self):
        ts = self.get_example_ts(4, 3)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)
        self.assertEqual(input_file.num_provenances, 1)
        timestamp = input_file.provenances_timestamp[0]
        iso = datetime.datetime.now().isoformat()
        self.assertEqual(timestamp.split("T")[0], iso.split("T")[0])
        record = input_file.provenances_record[0]
        self.assertEqual(record["software"], "tsinfer")
        a = list(input_file.provenances())
        self.assertEqual(len(a), 1)
        self.assertEqual(a[0][0], timestamp)
        self.assertEqual(a[0][1], record)

    def test_variant_errors(self):
        input_file = formats.SampleData(sequence_length=10)
        genotypes = np.zeros(2, np.uint8)
        input_file.add_site(0, alleles=["0", "1"], genotypes=genotypes)
        for bad_position in [-1, 10, 100]:
            self.assertRaises(
                ValueError, input_file.add_site, position=bad_position,
                alleles=["0", "1"], genotypes=genotypes)
        for bad_genotypes in [[0, 2], [-1, 0], [], [0], [0, 0, 0]]:
            genotypes = np.array(bad_genotypes, dtype=np.uint8)
            self.assertRaises(
                ValueError, input_file.add_site, position=1,
                alleles=["0", "1"], genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=1,
            alleles=["0", "1", "2"], genotypes=np.zeros(2, dtype=np.int8))
        self.assertRaises(
            ValueError, input_file.add_site, position=1,
            alleles=["0"], genotypes=np.array([0, 1], dtype=np.int8))
        self.assertRaises(
            ValueError, input_file.add_site, position=1,
            alleles=["0", "1"], genotypes=np.array([0, 2], dtype=np.int8))
        self.assertRaises(
            ValueError, input_file.add_site, position=1,
            alleles=["0", "0"], genotypes=np.array([0, 2], dtype=np.int8))

    def test_invalid_inference_sites(self):
        # Trying to add singletons or fixed sites as inference sites
        # raise and error
        input_file = formats.SampleData()
        # Make sure this is OK
        input_file.add_site(0, [0, 1, 1], inference=True)
        self.assertRaises(
            ValueError, input_file.add_site,
            position=1, genotypes=[0, 0, 0], inference=True)
        self.assertRaises(
            ValueError, input_file.add_site,
            position=1, genotypes=[1, 0, 0], inference=True)
        self.assertRaises(
            ValueError, input_file.add_site,
            position=1, genotypes=[1, 1, 1], inference=True)
        input_file.add_site(
            position=1, genotypes=[1, 0, 1], inference=True)

    def test_duplicate_sites(self):
        # Duplicate sites are not accepted.
        input_file = formats.SampleData()
        alleles = ["0", "1"]
        genotypes = [0, 1, 1]
        input_file.add_site(position=0, alleles=alleles, genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=0, alleles=alleles,
            genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=0, alleles=alleles,
            genotypes=genotypes)
        input_file.add_site(position=1, alleles=alleles, genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=1, alleles=alleles,
            genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=1, alleles=alleles,
            genotypes=genotypes)

    def test_unordered_sites(self):
        # Sites must be specified in sorted order.
        input_file = formats.SampleData()
        alleles = ["0", "1"]
        genotypes = [0, 1, 1]
        input_file.add_site(position=0, alleles=alleles, genotypes=genotypes)
        input_file.add_site(position=1, alleles=alleles, genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=0.5, alleles=alleles,
            genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=0.9999, alleles=alleles,
            genotypes=genotypes)
        input_file.add_site(position=2, alleles=alleles, genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=0.5, alleles=alleles,
            genotypes=genotypes)
        self.assertRaises(
            ValueError, input_file.add_site, position=1.88, alleles=alleles,
            genotypes=genotypes)

    def test_insufficient_samples(self):
        sample_data = formats.SampleData(sequence_length=10)
        self.assertRaises(
            ValueError, sample_data.add_site, position=0, alleles=["0", "1"],
            genotypes=[])
        sample_data = formats.SampleData(sequence_length=10)
        self.assertRaises(
            ValueError, sample_data.add_site, position=0, alleles=["0", "1"],
            genotypes=[0])
        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_individual(ploidy=3)
        self.assertRaises(
            ValueError, sample_data.add_site, position=0, alleles=["0", "1"],
            genotypes=[0])

    def test_add_population_errors(self):
        sample_data = formats.SampleData(sequence_length=10)
        self.assertRaises(
            TypeError, sample_data.add_population, metadata=234)

    def test_add_state_machine(self):
        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_individual()
        self.assertRaises(ValueError, sample_data.add_population)

        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_site(0.1, genotypes=[0, 1])
        self.assertRaises(ValueError, sample_data.add_population)
        self.assertRaises(ValueError, sample_data.add_individual)

    def test_add_population_return(self):
        sample_data = formats.SampleData(sequence_length=10)
        pid = sample_data.add_population({"a": 1})
        self.assertEqual(pid, 0)
        pid = sample_data.add_population()
        self.assertEqual(pid, 1)
        pid = sample_data.add_population()
        self.assertEqual(pid, 2)

    def test_population_metadata(self):
        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_population({"a": 1})
        sample_data.add_population({"b": 2})
        sample_data.add_individual(population=0)
        sample_data.add_individual(population=1)
        sample_data.add_site(position=0, genotypes=[0, 1])
        sample_data.finalise()

        self.assertEqual(sample_data.populations_metadata[0], {"a": 1})
        self.assertEqual(sample_data.populations_metadata[1], {"b": 2})
        self.assertEqual(sample_data.samples_population[0], 0)
        self.assertEqual(sample_data.samples_population[1], 1)

    def test_individual_metadata(self):
        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_population({"a": 1})
        sample_data.add_population({"b": 2})
        sample_data.add_individual(population=0)
        sample_data.add_individual(population=1)
        sample_data.finalise()
        self.assertEqual(sample_data.populations_metadata[0], {"a": 1})
        self.assertEqual(sample_data.populations_metadata[1], {"b": 2})

    def test_add_individual_return(self):
        sample_data = formats.SampleData(sequence_length=10)
        iid, sids = sample_data.add_individual()
        self.assertEqual(iid, 0)
        self.assertEqual(sids, [0])
        iid, sids = sample_data.add_individual(ploidy=1)
        self.assertEqual(iid, 1)
        self.assertEqual(sids, [1])
        iid, sids = sample_data.add_individual(ploidy=5)
        self.assertEqual(iid, 2)
        self.assertEqual(sids, [2, 3, 4, 5, 6])

    def test_add_individual_errors(self):
        sample_data = formats.SampleData(sequence_length=10)
        self.assertRaises(TypeError, sample_data.add_individual, metadata=234)
        self.assertRaises(ValueError, sample_data.add_individual, population=0)
        sample_data = formats.SampleData(sequence_length=10)
        sample_data.add_population()
        self.assertRaises(ValueError, sample_data.add_individual, population=1)
        self.assertRaises(ValueError, sample_data.add_individual, location="x234")
        self.assertRaises(ValueError, sample_data.add_individual, ploidy=0)

    def test_no_data(self):
        sample_data = formats.SampleData(sequence_length=10)
        self.assertRaises(ValueError, sample_data.finalise)

    def test_add_site_return(self):
        sample_data = formats.SampleData(sequence_length=10)
        sid = sample_data.add_site(0, [0, 1])
        self.assertEqual(sid, 0)
        sid = sample_data.add_site(1, [0, 1])
        self.assertEqual(sid, 1)

    def test_genotypes(self):
        ts = self.get_example_ts(13, 12)
        self.assertGreater(ts.num_sites, 1)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        for v in ts.variants():
            input_file.add_site(v.site.position, v.genotypes, v.alleles)
        input_file.finalise()

        self.assertLess(np.sum(input_file.sites_inference[:]), ts.num_sites)
        all_genotypes = input_file.genotypes()
        for v in ts.variants():
            site_id, g = next(all_genotypes)
            self.assertEqual(site_id, v.site.id)
            self.assertTrue(np.array_equal(g, v.genotypes))
        self.assertIsNone(next(all_genotypes, None), None)

        inference_genotypes = input_file.genotypes(inference_sites=True)
        non_inference_genotypes = input_file.genotypes(inference_sites=False)
        for v in ts.variants():
            freq = np.sum(v.genotypes)
            if 1 < freq < ts.num_samples:
                site_id, g = next(inference_genotypes)
                self.assertEqual(site_id, v.site.id)
                self.assertTrue(np.array_equal(g, v.genotypes))
            else:
                site_id, g = next(non_inference_genotypes)
                self.assertEqual(site_id, v.site.id)
                self.assertTrue(np.array_equal(g, v.genotypes))
        self.assertIsNone(next(inference_genotypes, None), None)

    def test_haplotypes(self):
        ts = self.get_example_ts(13, 12)
        self.assertGreater(ts.num_sites, 1)
        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        for v in ts.variants():
            input_file.add_site(v.site.position, v.genotypes, v.alleles)
        input_file.finalise()

        G = ts.genotype_matrix()
        j = 0
        for h in input_file.haplotypes():
            self.assertTrue(np.array_equal(h, G[:, j]))
            j += 1
        self.assertEqual(j, ts.num_samples)

        selection = input_file.sites_inference[:] == 1
        j = 0
        for h in input_file.haplotypes(inference_sites=True):
            self.assertTrue(np.array_equal(h, G[selection, j]))
            j += 1
        self.assertEqual(j, ts.num_samples)

        selection = input_file.sites_inference[:] == 0
        j = 0
        for h in input_file.haplotypes(inference_sites=False):
            self.assertTrue(np.array_equal(h, G[selection, j]))
            j += 1
        self.assertEqual(j, ts.num_samples)

    def test_invariant_sites(self):
        n = 10
        m = 10
        for value in [0, 1]:
            G = np.zeros((m, n), dtype=np.int8) + value
            input_file = formats.SampleData(sequence_length=m)
            for j in range(m):
                input_file.add_site(j, G[j])
            input_file.finalise()
            self.assertEqual(input_file.num_sites, m)
            self.assertTrue(
                np.all(input_file.sites_inference[:] == np.zeros(m, dtype=int)))

    def test_ts_with_invariant_sites(self):
        ts = self.get_example_ts(5, 3)
        t = ts.dump_tables()
        positions = set(site.position for site in ts.sites())
        for j in range(10):
            pos = 1 / (j + 1)
            if pos not in positions:
                t.sites.add_row(position=pos, ancestral_state="0")
                positions.add(pos)
        self.assertGreater(len(positions), ts.num_sites)
        t.sort()
        ts = t.tree_sequence()

        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)
        self.assertGreater(len(str(input_file)), 0)

    def test_ts_with_root_mutations(self):
        ts = self.get_example_ts(5, 3)
        t = ts.dump_tables()
        positions = set(site.position for site in ts.sites())
        for tree in ts.trees():
            pos = tree.interval[0]
            if pos not in positions:
                site_id = t.sites.add_row(position=pos, ancestral_state="0")
                t.mutations.add_row(site=site_id, node=tree.root, derived_state="1")
                positions.add(pos)
        self.assertGreater(len(positions), ts.num_sites)
        t.sort()
        ts = t.tree_sequence()

        input_file = formats.SampleData(sequence_length=ts.sequence_length)
        self.verify_data_round_trip(ts, input_file)

    def test_copy_error_wrong_mode(self):
        data = formats.SampleData()
        self.assertRaises(ValueError, data.copy)
        data = formats.SampleData()
        self.assertRaises(ValueError, data.copy)
        data = formats.SampleData()
        data.add_site(position=0, alleles=["0", "1"], genotypes=[0, 1])
        data.finalise()
        copy = data.copy()
        self.assertRaises(ValueError, copy.copy)

    def test_copy_new_uuid(self):
        data = formats.SampleData()
        data.add_site(position=0, alleles=["0", "1"], genotypes=[0, 1])
        data.finalise()
        copy = data.copy()
        copy.finalise()
        self.assertNotEqual(copy.uuid, data.uuid)
        self.assertTrue(copy.data_equal(data))

    def test_copy_update_inference_sites(self):
        with formats.SampleData() as data:
            for j in range(4):
                data.add_site(position=j, alleles=["0", "1"], genotypes=[0, 1, 1, 0])
        self.assertEqual(list(data.sites_inference), [1, 1, 1, 1])

        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "samples.tmp")
            for copy_path in [None, filename]:
                copy = data.copy(path=copy_path)
                copy.finalise()
                self.assertTrue(copy.data_equal(data))
                with data.copy(path=copy_path) as copy:
                    inference = [0, 1, 0, 1]
                    copy.sites_inference = inference
                self.assertFalse(copy.data_equal(data))
                self.assertEqual(list(copy.sites_inference), inference)
                self.assertEqual(list(data.sites_inference), [1, 1, 1, 1])

    def test_update_inference_sites_bad_data(self):
        def set_value(data, value):
            data.sites_inference = value

        data = formats.SampleData()
        for j in range(4):
            data.add_site(position=j, alleles=["0", "1"], genotypes=[0, 1, 1, 0])
        data.finalise()
        self.assertEqual(list(data.sites_inference), [1, 1, 1, 1])
        copy = data.copy()
        for bad_shape in [[], np.arange(100), np.zeros((2, 2))]:
            self.assertRaises(ValueError, set_value, copy, bad_shape)
        bad_data = [
            ["a", "b", "c", "d"], [2**10 for _ in range(4)], [0, 1, 0, 2],
            [0, 0, 0, -1]]
        for a in bad_data:
            self.assertRaises(ValueError, set_value, copy, a)

    def test_update_inference_sites_non_copy_mode(self):
        def set_value(data, value):
            data.sites_inference = value

        data = formats.SampleData()
        data.add_site(position=0, alleles=["0", "1"], genotypes=[0, 1, 1, 0])
        self.assertRaises(ValueError, set_value, data, [])
        data.finalise()
        self.assertRaises(ValueError, set_value, data, [])

    def test_overwrite_partial(self):
        # Check that we can correctly overwrite partially written and
        # unfinalised files. See
        # https://github.com/tskit-dev/tsinfer/issues/64
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "samples.tmp")
            for _ in range(10):
                data = formats.SampleData(path=filename)
                for j in range(10):
                    data.add_site(j, [0, 1])


class TestAncestorData(unittest.TestCase, DataContainerMixin):
    """
    Test cases for the sample data file format.
    """
    tested_class = formats.AncestorData

    def get_example_data(self, sample_size, sequence_length, num_ancestors):
        ts = msprime.simulate(
            sample_size, recombination_rate=1, mutation_rate=10,
            length=sequence_length, random_seed=100)
        sample_data = formats.SampleData(sequence_length=ts.sequence_length)
        for v in ts.variants():
            sample_data.add_site(v.site.position, v.genotypes, v.alleles)
        sample_data.finalise()

        num_sites = sample_data.num_inference_sites
        ancestors = []
        for j in range(num_ancestors):
            haplotype = np.zeros(num_sites, dtype=np.uint8) + tsinfer.UNKNOWN_ALLELE
            start = j
            end = max(num_sites - j, start + 1)
            self.assertLess(start, end)
            haplotype[start: end] = 0
            if start + j < end:
                haplotype[start + j: end] = 1
            self.assertTrue(np.all(haplotype[:start] == tsinfer.UNKNOWN_ALLELE))
            self.assertTrue(np.all(haplotype[end:] == tsinfer.UNKNOWN_ALLELE))
            focal_sites = np.array([start + k for k in range(j)], dtype=np.int32)
            focal_sites = focal_sites[focal_sites < end]
            haplotype[focal_sites] = 1
            ancestors.append((start, end, 2 * j + 1, focal_sites, haplotype))
        return sample_data, ancestors

    def verify_data_round_trip(self, sample_data, ancestor_data, ancestors):
        for start, end, time, focal_sites, haplotype in ancestors:
            ancestor_data.add_ancestor(start, end, time, focal_sites, haplotype)
        ancestor_data.finalise()

        self.assertGreater(len(ancestor_data.uuid), 0)
        self.assertEqual(ancestor_data.sample_data_uuid, sample_data.uuid)
        self.assertEqual(ancestor_data.format_name, formats.AncestorData.FORMAT_NAME)
        self.assertEqual(
            ancestor_data.format_version, formats.AncestorData.FORMAT_VERSION)
        self.assertEqual(ancestor_data.num_sites, sample_data.num_inference_sites)
        self.assertEqual(ancestor_data.num_ancestors, len(ancestors))

        ancestors_list = list(ancestor_data.ancestors())
        stored_start = ancestor_data.start[:]
        stored_end = ancestor_data.end[:]
        stored_time = ancestor_data.time[:]
        stored_ancestors = ancestor_data.ancestor[:]
        stored_focal_sites = ancestor_data.focal_sites[:]
        for j, (start, end, time, focal_sites, haplotype) in enumerate(ancestors):
            self.assertEqual(stored_start[j], start)
            self.assertEqual(stored_end[j], end)
            self.assertEqual(stored_time[j], time)
            self.assertTrue(np.array_equal(stored_focal_sites[j], focal_sites))
            self.assertTrue(np.array_equal(stored_ancestors[j], haplotype[start: end]))
            self.assertTrue(np.array_equal(ancestors_list[j], haplotype[start: end]))

    def test_defaults_no_path(self):
        sample_data, ancestors = self.get_example_data(10, 10, 40)
        ancestor_data = tsinfer.AncestorData(sample_data)
        self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
        for _, array in ancestor_data.arrays():
            self.assertEqual(array.compressor, None)

    def test_defaults_with_path(self):
        sample_data, ancestors = self.get_example_data(10, 10, 40)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "ancestors.tmp")
            ancestor_data = tsinfer.AncestorData(sample_data, path=filename)
            self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
            compressor = formats.DEFAULT_COMPRESSOR
            for _, array in ancestor_data.arrays():
                self.assertEqual(array.compressor, compressor)
            with tsinfer.load(filename) as other:
                self.assertEqual(other, ancestor_data)

    def test_provenance(self):
        sample_data, ancestors = self.get_example_data(10, 10, 40)
        ancestor_data = tsinfer.AncestorData(sample_data)
        self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
        self.assertEqual(ancestor_data.num_provenances, 1)
        timestamp = ancestor_data.provenances_timestamp[0]
        iso = datetime.datetime.now().isoformat()
        self.assertEqual(timestamp.split("T")[0], iso.split("T")[0])
        record = ancestor_data.provenances_record[0]
        self.assertEqual(record["software"], "tsinfer")
        a = list(ancestor_data.provenances())
        self.assertEqual(len(a), 1)
        self.assertEqual(a[0][0], timestamp)
        self.assertEqual(a[0][1], record)

    def test_chunk_size(self):
        N = 20
        for chunk_size in [1, 2, 3, N - 1, N, N + 1]:
            sample_data, ancestors = self.get_example_data(6, 1, N)
            self.assertGreater(sample_data.num_inference_sites, 2 * N)
            ancestor_data = tsinfer.AncestorData(
                sample_data, chunk_size=chunk_size)
            self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
            self.assertEqual(ancestor_data.ancestor.chunks, (chunk_size,))
            self.assertEqual(ancestor_data.focal_sites.chunks, (chunk_size,))
            self.assertEqual(ancestor_data.start.chunks, (chunk_size,))
            self.assertEqual(ancestor_data.end.chunks, (chunk_size,))
            self.assertEqual(ancestor_data.time.chunks, (chunk_size,))

    def test_filename(self):
        sample_data, ancestors = self.get_example_data(10, 2, 40)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            filename = os.path.join(tempdir, "ancestors.tmp")
            ancestor_data = tsinfer.AncestorData(
                sample_data, path=filename)
            self.assertTrue(os.path.exists(filename))
            self.assertFalse(os.path.isdir(filename))
            self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
            other_ancestor_data = formats.AncestorData.load(filename)
            self.assertIsNot(other_ancestor_data, ancestor_data)
            self.assertEqual(other_ancestor_data, ancestor_data)

    def test_chunk_size_file_equal(self):
        N = 60
        sample_data, ancestors = self.get_example_data(22, 16, N)
        with tempfile.TemporaryDirectory(prefix="tsinf_format_test") as tempdir:
            files = []
            for chunk_size in [5, 7]:
                filename = os.path.join(tempdir, "samples_{}.tmp".format(chunk_size))
                files.append(filename)
                ancestor_data = tsinfer.AncestorData(
                    sample_data, path=filename, chunk_size=chunk_size)
                self.verify_data_round_trip(sample_data, ancestor_data, ancestors)
                self.assertEqual(ancestor_data.ancestor.chunks, (chunk_size,))
            # Now reload the files and check they are equal
            file0 = formats.AncestorData.load(files[0])
            file1 = formats.AncestorData.load(files[1])
            self.assertTrue(file0.data_equal(file1))

    def test_add_ancestor_errors(self):
        sample_data, ancestors = self.get_example_data(22, 16, 30)
        ancestor_data = tsinfer.AncestorData(sample_data)
        num_sites = ancestor_data.num_sites
        haplotype = np.zeros(num_sites, dtype=np.int8)
        ancestor_data.add_ancestor(
            start=0, end=num_sites, time=1, focal_sites=[], haplotype=haplotype)
        for bad_start in [-1, -100, num_sites, num_sites + 1]:
            self.assertRaises(
                ValueError, ancestor_data.add_ancestor,
                start=bad_start, end=num_sites, time=0, focal_sites=[],
                haplotype=haplotype)
        for bad_end in [-1, 0, num_sites + 1, 10 * num_sites]:
            self.assertRaises(
                ValueError, ancestor_data.add_ancestor,
                start=0, end=bad_end, time=1, focal_sites=[], haplotype=haplotype)
        for bad_time in [-1, 0]:
            self.assertRaises(
                ValueError, ancestor_data.add_ancestor,
                start=0, end=num_sites, time=bad_time, focal_sites=[],
                haplotype=haplotype)
        self.assertRaises(
            ValueError, ancestor_data.add_ancestor,
            start=0, end=num_sites, time=1, focal_sites=[],
            haplotype=np.zeros(num_sites + 1, dtype=np.uint8))
        # Haplotypes must be < 2
        self.assertRaises(
            ValueError, ancestor_data.add_ancestor,
            start=0, end=num_sites, time=1, focal_sites=[],
            haplotype=np.zeros(num_sites, dtype=np.uint8) + 2)

        # focal sites must be within start:end
        self.assertRaises(
            ValueError, ancestor_data.add_ancestor,
            start=1, end=num_sites, time=1, focal_sites=[0],
            haplotype=np.ones(num_sites, dtype=np.uint8))
        self.assertRaises(
            ValueError, ancestor_data.add_ancestor,
            start=0, end=num_sites - 2, time=1, focal_sites=[num_sites - 1],
            haplotype=np.ones(num_sites, dtype=np.uint8))
        # focal sites must be set to 1
        self.assertRaises(
            ValueError, ancestor_data.add_ancestor,
            start=0, end=num_sites, time=1, focal_sites=[0],
            haplotype=np.zeros(num_sites, dtype=np.uint8))


class BufferedItemWriterMixin(object):
    """
    Tests to ensure that the buffered item writer works as expected.
    """
    def filter_warnings_verify_round_trip(self, source):
        # Zarr currently emits an error when dealing with object arrays.
        # https://github.com/zarr-developers/zarr/issues/257
        # As a workaround, we filter this warning.
        # This should be removed when the bug has been fixed upstream.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.verify_round_trip(source)

    def verify_round_trip(self, source):
        """
        Verify that we can round trip the specified mapping of arrays
        using the buffered item writer.
        """
        # Create a set of empty arrays like the originals.
        dest = {}
        num_rows = -1
        for key, array in source.items():
            dest[key] = zarr.empty_like(array)
            if num_rows == -1:
                num_rows = array.shape[0]
            assert num_rows == array.shape[0]
        assert num_rows != -1
        writer = formats.BufferedItemWriter(dest, num_threads=self.num_threads)
        for j in range(num_rows):
            row = {key: array[j] for key, array in source.items()}
            writer.add(**row)
        writer.flush()

        for key, source_array in source.items():
            dest_array = dest[key]
            if source_array.dtype.str == "|O":
                # Object arrays have to be treated differently.
                self.assertTrue(source_array.shape == dest_array.shape)
                for a, b in zip(source_array, dest_array):
                    if isinstance(a, np.ndarray):
                        self.assertTrue(np.array_equal(a, b))
                    else:
                        self.assertEqual(a, b)
            else:
                self.assertTrue(np.array_equal(source_array[:], dest_array[:]))
            self.assertEqual(source_array.chunks, dest_array.chunks)
        return dest

    def test_one_array(self):
        self.verify_round_trip({"a": zarr.ones(10)})

    def test_two_arrays(self):
        self.verify_round_trip({"a": zarr.ones(10), "b": zarr.zeros(10)})

    def verify_dtypes(self, chunk_size=None):
        n = 100
        if chunk_size is None:
            chunk_size = 100
        dtypes = [np.int8, np.uint8, np.int32, np.uint32, np.float64, np.float32]
        source = {
            str(dtype): zarr.array(np.arange(n, dtype=dtype), chunks=(chunk_size,))
            for dtype in dtypes}
        dest = self.verify_round_trip(source)
        for dtype in dtypes:
            self.assertEqual(dest[str(dtype)].dtype, dtype)

    def test_mixed_dtypes(self):
        self.verify_dtypes()

    def test_mixed_dtypes_chunk_size_1(self):
        self.verify_dtypes(1)

    def test_mixed_dtypes_chunk_size_2(self):
        self.verify_dtypes(2)

    def test_mixed_dtypes_chunk_size_3(self):
        self.verify_dtypes(3)

    def test_mixed_dtypes_chunk_size_10000(self):
        self.verify_dtypes(10000)

    def test_2d_array(self):
        a = zarr.array(np.arange(100).reshape((10, 10)))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_1_1(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(1, 1))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_1_2(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(1, 2))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_2_1(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(1, 2))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_1_100(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(1, 100))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_100_1(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(100, 1))
        self.verify_round_trip({"a": a})

    def test_2d_array_chunk_size_10_10(self):
        a = zarr.array(np.arange(100).reshape((10, 10)), chunks=(5, 10))
        self.verify_round_trip({"a": a})

    def test_3d_array(self):
        a = zarr.array(np.arange(27).reshape((3, 3, 3)))
        self.verify_round_trip({"a": a})

    def test_3d_array_chunks_size_1_1_1(self):
        a = zarr.array(np.arange(27).reshape((3, 3, 3)), chunks=(1, 1, 1))
        self.verify_round_trip({"a": a})

    def test_ragged_array_int32(self):
        n = 10
        z = zarr.empty(n, dtype="array:i4")
        for j in range(n):
            z[j] = np.arange(j)
        self.filter_warnings_verify_round_trip({"z": z})

    def test_square_object_array_int32(self):
        n = 10
        z = zarr.empty(n, dtype="array:i4")
        for j in range(n):
            z[j] = np.arange(n)
        self.filter_warnings_verify_round_trip({"z": z})

    def test_json_object_array(self):
        for chunks in [1, 2, 5, 10, 100]:
            n = 10
            z = zarr.empty(
                n, dtype=object, object_codec=formats.TempJSON(), chunks=(chunks,))
            for j in range(n):
                z[j] = {str(k): k for k in range(j)}
            self.filter_warnings_verify_round_trip({"z": z})

    def test_empty_string_list(self):
        z = zarr.empty(1, dtype=object, object_codec=formats.TempJSON(), chunks=(2,))
        z[0] = ["", ""]
        self.filter_warnings_verify_round_trip({"z": z})

    def test_mixed_chunk_sizes(self):
        source = {"a": zarr.zeros(10, chunks=(1,)), "b": zarr.zeros(10, chunks=(2,))}
        self.assertRaises(ValueError, formats.BufferedItemWriter, source)


class TestBufferedItemWriterSynchronous(unittest.TestCase, BufferedItemWriterMixin):
    num_threads = 0


class TestBufferedItemWriterThreads1(unittest.TestCase, BufferedItemWriterMixin):
    num_threads = 1


class TestBufferedItemWriterThreads2(unittest.TestCase, BufferedItemWriterMixin):
    num_threads = 2


class TestBufferedItemWriterThreads20(unittest.TestCase, BufferedItemWriterMixin):
    num_threads = 20



import os
import msprime
import sys
sys.path.insert(0, os.path.abspath('..'))
import tsinfer


ts = msprime.simulate(5, mutation_rate=0.7, random_seed=10)
tree = ts.first()
print(ts.num_sites)
print(tree.draw(format="unicode"))

with tsinfer.SampleData(path="toy.samples") as sample_data:
    sample_data.add_site(10, [0, 1, 0, 0, 0], ["A", "T"])
    sample_data.add_site(12, [0, 0, 0, 1, 1], ["G", "C"])
    sample_data.add_site(23, [0, 1, 1, 0, 0], ["C", "A"])
    sample_data.add_site(37, [0, 1, 1, 0, 0], ["G", "C"])
    sample_data.add_site(40, [0, 0, 0, 1, 1], ["A", "C"])
    sample_data.add_site(50, [0, 1, 0, 0, 0], ["T", "G"])

print(sample_data)

inferred_ts = tsinfer.infer(sample_data)
for tree in inferred_ts.trees():
    print(tree.draw(format="unicode"))

for sample_id, h in enumerate(inferred_ts.haplotypes()):
    print(sample_id, h, sep="\t")



#
# Copyright (C) 2018 University of Oxford
#
# This file is part of tsinfer.
#
# tsinfer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tsinfer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with tsinfer.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Command line interfaces to tsinfer.
"""
import argparse
import os
import os.path
import logging
import resource

import daiquiri
import msprime
import tqdm
import humanize
import time

import tsinfer


logger = logging.getLogger(__name__)


class ProgressMonitor(object):
    """
    Class responsible for managing in the tqdm progress monitors.
    """
    def __init__(
            self, enabled=True, generate_ancestors=False, match_ancestors=False,
            match_samples=False):
        self.enabled = enabled
        self.num_bars = 0
        if generate_ancestors:
            self.num_bars += 2
        if match_ancestors:
            self.num_bars += 1
        if match_samples:
            self.num_bars += 3
        self.current_count = 0
        self.current_instance = None
        # Only show extra detail if we are runing match-ancestors by itself.
        self.show_detail = self.num_bars == 1
        self.descriptions = {
            "ga_add_sites": "ga-add",
            "ga_generate": "ga-gen",
            "ma_match": "ma-match",
            "ms_match": "ms-match",
            "ms_paths": "ms-paths",
            "ms_sites": "ms-sites",
        }

    def set_detail(self, info):
        if self.show_detail:
            self.current_instance.set_postfix(info)

    def get(self, key, total):
        self.current_count += 1
        desc = "{:<8} ({}/{})".format(
            self.descriptions[key], self.current_count, self.num_bars)
        bar_format = (
            "{desc}{percentage:3.0f}%|{bar}"
            "| {n_fmt}/{total_fmt} [{elapsed}, {rate_fmt}{postfix}]")
        self.current_instance = tqdm.tqdm(
            desc=desc, total=total, disable=not self.enabled,
            bar_format=bar_format, dynamic_ncols=True, smoothing=0.01,
            unit_scale=True)
        return self.current_instance


__before = time.time()


def summarise_usage():
    wall_time = humanize.naturaldelta(time.time() - __before)
    rusage = resource.getrusage(resource.RUSAGE_SELF)
    user_time = humanize.naturaldelta(rusage.ru_utime)
    sys_time = rusage.ru_stime
    max_rss = humanize.naturalsize(rusage.ru_maxrss * 1024, binary=True)
    logger.info("wall time = {}".format(wall_time))
    logger.info("rusage: user={}; sys={:.2f}s; max_rss={}".format(
        user_time, sys_time, max_rss))


def get_default_path(path, input_path, extension):
    if path is None:
        path = os.path.splitext(input_path)[0] + extension
    return path


def get_ancestors_path(path, input_path):
    return get_default_path(path, input_path, ".ancestors")


def get_ancestors_trees_path(path, input_path):
    return get_default_path(path, input_path, ".ancestors.trees")


def get_output_trees_path(path, input_path):
    return get_default_path(path, input_path, ".trees")


def setup_logging(args):
    log_level = "WARN"
    if args.verbosity > 0:
        log_level = "INFO"
    if args.verbosity > 1:
        log_level = "DEBUG"
    if args.log_section is None:
        daiquiri.setup(level=log_level)
    else:
        daiquiri.setup(level="WARN")
        logger = logging.getLogger(args.log_section)
        logger.setLevel(log_level)


def summarise_tree_sequence(path, ts):
    print("path      =", path)
    print("size      =", humanize.naturalsize(os.path.getsize(path), binary=True))
    print("edges     =", ts.num_edges)
    print("trees     =", ts.num_trees)
    print("sites     =", ts.num_sites)
    print("mutations =", ts.num_mutations)
    # TODO Add optional tree statistics like mean degree, etc.


def run_list(args):
    setup_logging(args)
    # First try to load with msprime.
    ts = None
    try:
        ts = msprime.load(args.path)
    except msprime.FileFormatError:
        pass
    if ts is None:
        tsinfer_file = tsinfer.load(args.path)
        if args.storage:
            print(tsinfer_file.info)
        else:
            print(tsinfer_file)
    else:
        summarise_tree_sequence(args.path, ts)


def run_infer(args):
    setup_logging(args)
    progress_monitor = ProgressMonitor(
        enabled=args.progress, generate_ancestors=True, match_ancestors=True,
        match_samples=True)
    sample_data = tsinfer.SampleData.load(args.input)
    ts = tsinfer.infer(
        sample_data, progress_monitor=progress_monitor, num_threads=args.num_threads)
    output_trees = get_output_trees_path(args.output_trees, args.input)
    logger.info("Writing output tree sequence to {}".format(output_trees))
    ts.dump(output_trees)
    summarise_usage()


def run_generate_ancestors(args):
    setup_logging(args)
    ancestors_path = get_ancestors_path(args.ancestors, args.input)
    progress_monitor = ProgressMonitor(enabled=args.progress, generate_ancestors=True)
    sample_data = tsinfer.SampleData.load(args.input)
    tsinfer.generate_ancestors(
        sample_data, progress_monitor=progress_monitor, path=ancestors_path,
        num_flush_threads=args.num_threads)
    summarise_usage()


def run_match_ancestors(args):
    setup_logging(args)
    ancestors_path = get_ancestors_path(args.ancestors, args.input)
    logger.info("Loading ancestral haplotypes from {}".format(ancestors_path))
    ancestors_trees = get_ancestors_trees_path(args.ancestors_trees, args.input)
    sample_data = tsinfer.SampleData.load(args.input)
    ancestor_data = tsinfer.AncestorData.load(ancestors_path)
    progress_monitor = ProgressMonitor(enabled=args.progress, match_ancestors=True)
    ts = tsinfer.match_ancestors(
        sample_data, ancestor_data,
        num_threads=args.num_threads, progress_monitor=progress_monitor,
        path_compression=not args.no_path_compression)
    logger.info("Writing ancestors tree sequence to {}".format(ancestors_trees))
    ts.dump(ancestors_trees)
    summarise_usage()


def run_match_samples(args):
    setup_logging(args)

    sample_data = tsinfer.SampleData.load(args.input)
    ancestors_trees = get_ancestors_trees_path(args.ancestors_trees, args.input)
    output_trees = get_output_trees_path(args.output_trees, args.input)
    logger.info("Loading ancestral genealogies from {}".format(ancestors_trees))
    ancestors_trees = msprime.load(ancestors_trees)
    progress_monitor = ProgressMonitor(enabled=args.progress, match_samples=True)
    ts = tsinfer.match_samples(
        sample_data, ancestors_trees, num_threads=args.num_threads,
        path_compression=not args.no_path_compression,
        simplify=not args.no_simplify,
        progress_monitor=progress_monitor)
    logger.info("Writing output tree sequence to {}".format(output_trees))
    ts.dump(output_trees)
    summarise_usage()


# def run_verify(args):
#     setup_logging(args)
#     print("FIXME!!!")
#     sys.exit(1)


def add_input_file_argument(parser):
    parser.add_argument(
        "input",
        help=(
            "The input sample data in tsinfer 'samples' format. Please see the "
            "documentation at http://tsinfer.readthedocs.io/ for information on "
            "how to import data into this format."))


def add_ancestors_file_argument(parser):
    parser.add_argument(
        "-a", "--ancestors", default=None,
        help=(
            "The path to the ancestor data file in tsinfer 'ancestors' format. "
            "If not specified, this defaults to the input samples file stem "
            "with the extension '.ancestors'. For example, if '1kg-chr1.samples' "
            "is the input file then the default ancestors file would be "
            "'1kg-chr1.ancestors'"))


def add_ancestors_trees_argument(parser):
    parser.add_argument(
        "-A", "--ancestors-trees", default=None,
        help=(
            "The path to the ancestor trees file in tskit '.trees' format. "
            "If not specified, this defaults to the input samples file stem "
            "with the extension '.ancestors.trees'. For example, if '1kg-chr1.samples' "
            "is the input file then the default ancestors file would be "
            "'1kg-chr1.ancestors.trees'"))


def add_output_trees_argument(parser):
    parser.add_argument(
        "-O", "--output-trees", default=None,
        help=(
            "The path to the output trees file in tskit '.trees' format. "
            "If not specified, this defaults to the input samples file stem "
            "with the extension '.trees'. For example, if '1kg-chr1.samples' "
            "is the input file then the default output file would be "
            "'1kg-chr1.trees'"))


def add_progress_argument(parser):
    parser.add_argument(
        "--progress", "-p", action="store_true",
        help="Show a progress monitor.")


def add_path_compression_argument(parser):
    parser.add_argument(
        "--no-path-compression", action="store_true",
        help="Disable path compression")


def add_simplify_argument(parser):
    parser.add_argument(
        "--no-simplify", action="store_true",
        help="Do not simplify the output tree sequence")


def add_logging_arguments(parser):
    log_sections = ["tsinfer.inference", "tsinfer.formats", "tsinfer.threads"]
    parser.add_argument(
        "-v", "--verbosity", action='count', default=0,
        help="Increase the verbosity")
    parser.add_argument(
        "--log-section", "-L", choices=log_sections, default=None,
        help=("Log messages only for the specified module"))


def add_num_threads_argument(parser):
    parser.add_argument(
        "--num-threads", "-t", type=int, default=0,
        help=(
            "The number of worker threads to use. If < 1, use a simpler unthreaded "
            "algorithm (default)."))


def get_cli_parser():
    top_parser = argparse.ArgumentParser(
        description="Command line interface for tsinfer.")
    top_parser.add_argument(
        "-V", "--version", action='version',
        version='%(prog)s {}'.format(tsinfer.__version__))

    subparsers = top_parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    parser = subparsers.add_parser(
        "generate-ancestors",
        aliases=["ga"],
        help=(
            "Generates a set of ancestors from the input sample data and stores "
            "the results in a tsinfer ancestors file."))
    add_input_file_argument(parser)
    add_ancestors_file_argument(parser)
    add_num_threads_argument(parser)
    add_progress_argument(parser)
    add_logging_arguments(parser)
    parser.set_defaults(runner=run_generate_ancestors)

    parser = subparsers.add_parser(
        "match-ancestors",
        aliases=["ma"],
        help=(
            "Matches the ancestors built by the 'generate-ancestors' command against "
            "each other using the model information specified in the input file "
            "and writes the output to a tskit .trees file."))
    add_input_file_argument(parser)
    add_logging_arguments(parser)
    add_ancestors_file_argument(parser)
    add_ancestors_trees_argument(parser)
    add_num_threads_argument(parser)
    add_progress_argument(parser)
    add_path_compression_argument(parser)
    parser.set_defaults(runner=run_match_ancestors)

    parser = subparsers.add_parser(
        "match-samples",
        aliases=["ms"],
        help=(
            "Matches the samples against the tree sequence structure built "
            "by the match-ancestors command"))
    add_input_file_argument(parser)
    add_logging_arguments(parser)
    add_ancestors_trees_argument(parser)
    add_path_compression_argument(parser)
    add_simplify_argument(parser)
    add_output_trees_argument(parser)
    add_num_threads_argument(parser)
    add_progress_argument(parser)
    parser.set_defaults(runner=run_match_samples)

    parser = subparsers.add_parser(
        "infer",
        help=(
            "Runs the generate-ancestors, match-ancestors and match-samples "
            "commands without writing the intermediate files to disk. Not "
            "recommended for large inferences."))
    add_input_file_argument(parser)
    add_logging_arguments(parser)
    add_output_trees_argument(parser)
    add_num_threads_argument(parser)
    add_progress_argument(parser)
    parser.set_defaults(runner=run_infer)

    parser = subparsers.add_parser(
        "list",
        aliases=["ls"],
        help=(
            "Show a summary of the specified tsinfer related file."))
    add_logging_arguments(parser)
    parser.add_argument(
        "path", help="The tsinfer file to show information about.")
    parser.add_argument(
        "--storage", "-s", action="store_true",
        help="Show detailed information about data storage.")
    parser.set_defaults(runner=run_list)

    return top_parser


def tsinfer_main(arg_list=None):
    parser = get_cli_parser()
    args = parser.parse_args(arg_list)
    args.runner(args)




import random
import os
import h5py
import zarr
import sys
import pandas as pd
import daiquiri
#import bsddb3
import time
import scipy
import pickle
import collections
import itertools
import tqdm
import shutil
import pprint
import numpy as np
import json

import matplotlib as mp
# Force matplotlib to not use any Xwindows backend.
mp.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns

import tsinfer
import msprime



def plot_breakpoints(ts, map_file, output_file):
    # Read in the recombination map using the read_hapmap engine,
    recomb_map = msprime.RecombinationMap.read_hapmap(map_file)

    # Now we get the positions and rates from the recombination
    # map and plot these using 500 bins.
    positions = np.array(recomb_map.get_positions()[1:])
    rates = np.array(recomb_map.get_rates()[1:])
    num_bins = 500
    v, bin_edges, _ = scipy.stats.binned_statistic(
        positions, rates, bins=num_bins)
    x = bin_edges[:-1][np.logical_not(np.isnan(v))]
    y = v[np.logical_not(np.isnan(v))]
    fig, ax1 = plt.subplots(figsize=(16, 6))
    ax1.plot(x, y, color="blue", label="Recombination rate")
    ax1.set_ylabel("Recombination rate")
    ax1.set_xlabel("Chromosome position")

    # Now plot the density of breakpoints along the chromosome
    breakpoints = np.array(list(ts.breakpoints()))
    ax2 = ax1.twinx()
    v, bin_edges = np.histogram(breakpoints, num_bins, density=True)
    ax2.plot(bin_edges[:-1], v, color="green", label="Breakpoint density")
    ax2.set_ylabel("Breakpoint density")
    ax2.set_xlim(1.5e7, 5.3e7)
    plt.legend()
    fig.savefig(output_file)


def make_errors(v, p):
    """
    For each sample an error occurs with probability p. Errors are generated by
    sampling values from the stationary distribution, that is, if we have an
    allele frequency of f, a 1 is emitted with probability f and a
    0 with probability 1 - f. Thus, there is a possibility that an 'error'
    will in fact result in the same value.
    """
    w = np.copy(v)
    if p > 0:
        m = v.shape[0]
        frequency = np.sum(v) / m
        # Randomly choose samples with probability p
        samples = np.where(np.random.random(m) < p)[0]
        # Generate observations from the stationary distribution.
        errors = (np.random.random(samples.shape[0]) < frequency).astype(int)
        w[samples] = errors
    return w


def generate_samples(ts, error_p):
    """
    Returns samples with a bits flipped with a specified probability.

    Rejects any variants that result in a fixed column.
    """
    S = np.zeros((ts.sample_size, ts.num_mutations), dtype=np.int8)
    for variant in ts.variants():
        done = False
        # Reject any columns that have no 1s or no zeros
        while not done:
            S[:, variant.index] = make_errors(variant.genotypes, error_p)
            s = np.sum(S[:, variant.index])
            done = 0 < s < ts.sample_size
    return S.T


def tsinfer_dev(
        n, L, seed, num_threads=1, recombination_rate=1e-8,
        error_rate=0, engine="C", log_level="WARNING",
        debug=True, progress=False, path_compression=True):

    np.random.seed(seed)
    random.seed(seed)
    L_megabases = int(L * 10**6)

    # daiquiri.setup(level=log_level)

    ts = msprime.simulate(
            n, Ne=10**4, length=L_megabases,
            recombination_rate=recombination_rate, mutation_rate=1e-8,
            random_seed=seed)
    if debug:
        print("num_sites = ", ts.num_sites)
    assert ts.num_sites > 0

    G = generate_samples(ts, error_rate)
    with tsinfer.SampleData(
            sequence_length=ts.sequence_length,
            path="tmp.samples",
            chunk_size=10) as sample_data:
        for j in range(ts.num_samples // 2):
            sample_data.add_individual(ploidy=2, metadata={"name": "ind_{}".format(j)})
        for site, genotypes in zip(ts.sites(), G):
            sample_data.add_site(
                site.position, genotypes,
                inference=np.sum(genotypes) > 1 and site.id % 2 == 0)

    ts = tsinfer.infer(sample_data, simplify=False)
    for tree in ts.trees():
        print(tree.draw(format="unicode"))

    # print("sample_data")
    # print(sample_data)
    # print(sample_data.samples_individual[:])
    # # sample_data.close()


    # with sample_data.copy() as other_data:
    #     print("other_data = ")
    #     print(other_data)
    #     print("mode = ", other_data._mode)


    # print("after")
    # print(other_data)
    # print("mode = ", other_data._mode)

    # # print(sample_data)

    # # sample_data.finalise()
    # # print(sample_data)
    # print("LOADED")
    # with tsinfer.SampleData.load("tmp.samples") as sample_data:
    #     print(sample_data)
    # print(sample_data)

    ancestor_data = tsinfer.generate_ancestors(sample_data, engine=engine)

    ancestors_ts = tsinfer.match_ancestors(sample_data, ancestor_data, engine=engine)
    output_ts = tsinfer.match_samples(sample_data, ancestors_ts, engine=engine)
    # dump_provenance(output_ts)


def dump_provenance(ts):
    print("dump provenance")
    for p in ts.provenances():
        print("-" * 50)
        print(p.timestamp)
        pprint.pprint(json.loads(p.record))


def build_profile_inputs(n, num_megabases):
    L = num_megabases * 10**6
    input_file = "tmp__NOBACKUP__/profile-n={}-m={}.input.hdf5".format(
            n, num_megabases)
    if os.path.exists(input_file):
        ts = msprime.load(input_file)
    else:
        ts = msprime.simulate(
            n, length=L, Ne=10**4, recombination_rate=1e-8, mutation_rate=1e-8,
            random_seed=10)
        print("Ran simulation: n = ", n, " num_sites = ", ts.num_sites,
                "num_trees =", ts.num_trees)
        ts.dump(input_file)
    filename = "tmp__NOBACKUP__/profile-n={}-m={}.samples".format(n, num_megabases)
    if os.path.exists(filename):
        os.unlink(filename)
    # daiquiri.setup(level="DEBUG")
    sample_data = tsinfer.SampleData.initialise(
        sequence_length=ts.sequence_length, path=filename, num_flush_threads=4)
    progress_monitor = tqdm.tqdm(total=ts.num_samples)
    for j in range(ts.num_samples):
        sample_data.add_sample(metadata={"name": "sample_{}".format(j)})
        progress_monitor.update()
    progress_monitor.close()
    progress_monitor = tqdm.tqdm(total=ts.num_sites)
    for variant in ts.variants():
        sample_data.add_site(
            variant.site.position, variant.alleles, variant.genotypes)
        progress_monitor.update()
    sample_data.finalise()
    progress_monitor.close()

    print(sample_data)

#     filename = "tmp__NOBACKUP__/profile-n={}_m={}.ancestors".format(n, num_megabases)
#     if os.path.exists(filename):
#         os.unlink(filename)
#     ancestor_data = tsinfer.AncestorData.initialise(sample_data, filename=filename)
#     tsinfer.build_ancestors(sample_data, ancestor_data, progress=True)
#     ancestor_data.finalise()

def copy_1kg():
    source = "tmp__NOBACKUP__/1kg_chr22.samples"
    sample_data = tsinfer.SampleData.load(source)
    copy = sample_data.copy("tmp__NOBACKUP__/1kg_chr22_copy.samples")
    copy.finalise()
    print(sample_data)
    print("copy = ")
    print(copy)

def tutorial_samples():
    import tqdm
    import msprime
    import tsinfer

    ts = msprime.simulate(
        sample_size=10000, Ne=10**4, recombination_rate=1e-8,
        mutation_rate=1e-8, length=10*10**6, random_seed=42)
    ts.dump("tmp__NOBACKUP__/simulation-source.trees")
    print("simulation done:", ts.num_trees, "trees and", ts.num_sites,  "sites")

    progress = tqdm.tqdm(total=ts.num_sites)
    with tsinfer.SampleData(
            path="tmp__NOBACKUP__/simulation.samples",
            sequence_length=ts.sequence_length,
            num_flush_threads=2) as sample_data:
        for var in ts.variants():
            sample_data.add_site(var.site.position, var.genotypes, var.alleles)
            progress.update()
    progress.close()

if __name__ == "__main__":

    np.set_printoptions(linewidth=20000)
    np.set_printoptions(threshold=20000000)

    # tutorial_samples()

    # build_profile_inputs(10, 1)
    # build_profile_inputs(100, 10)
    # build_profile_inputs(1000, 100)
    # build_profile_inputs(10**4, 100)
    # build_profile_inputs(10**5, 100)

    # for j in range(1, 100):
    #     tsinfer_dev(15, 0.5, seed=j, num_threads=0, engine="P", recombination_rate=1e-8)
    # copy_1kg()
    # tsinfer_dev(6, 0.3, seed=1, num_threads=0, engine="C", recombination_rate=1e-8)



#     for seed in range(1, 10000):
#         print(seed)
#         # tsinfer_dev(40, 2.5, seed=seed, num_threads=1, genotype_quality=1e-3, engine="C")
